#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import ipaddress

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine.clients import progress
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import port as neutron_port
from heat.engine.resources.openstack.neutron import subnet
from heat.engine.resources.openstack.nova import server_network_mixin
from heat.engine.resources import scheduler_hints as sh
from heat.engine.resources import server_base
from heat.engine import support
from heat.engine import translation
from heat.rpc import api as rpc_api

cfg.CONF.import_opt('default_software_config_transport', 'heat.common.config')
cfg.CONF.import_opt('default_user_data_format', 'heat.common.config')

LOG = logging.getLogger(__name__)

NOVA_MICROVERSIONS = (MICROVERSION_TAGS, MICROVERSION_STR_NETWORK,
                      MICROVERSION_NIC_TAGS, MICROVERSION_PERSONALITY_REMOVED
                      ) = ('2.26', '2.37', '2.42', '2.57')


class Server(server_base.BaseServer, sh.SchedulerHintsMixin,
             server_network_mixin.ServerNetworkMixin):
    """A resource for managing Nova instances.

    A Server resource manages the running virtual machine instance within an
    OpenStack cloud.
    """

    PROPERTIES = (
        NAME, IMAGE, BLOCK_DEVICE_MAPPING, BLOCK_DEVICE_MAPPING_V2,
        FLAVOR, FLAVOR_UPDATE_POLICY, IMAGE_UPDATE_POLICY, KEY_NAME,
        ADMIN_USER, AVAILABILITY_ZONE, SECURITY_GROUPS, NETWORKS,
        SCHEDULER_HINTS, METADATA, USER_DATA_FORMAT, USER_DATA,
        RESERVATION_ID, CONFIG_DRIVE, DISK_CONFIG, PERSONALITY,
        ADMIN_PASS, SOFTWARE_CONFIG_TRANSPORT, USER_DATA_UPDATE_POLICY,
        TAGS, DEPLOYMENT_SWIFT_DATA
    ) = (
        'name', 'image', 'block_device_mapping', 'block_device_mapping_v2',
        'flavor', 'flavor_update_policy', 'image_update_policy', 'key_name',
        'admin_user', 'availability_zone', 'security_groups', 'networks',
        'scheduler_hints', 'metadata', 'user_data_format', 'user_data',
        'reservation_id', 'config_drive', 'diskConfig', 'personality',
        'admin_pass', 'software_config_transport', 'user_data_update_policy',
        'tags', 'deployment_swift_data'
    )

    _BLOCK_DEVICE_MAPPING_KEYS = (
        BLOCK_DEVICE_MAPPING_DEVICE_NAME, BLOCK_DEVICE_MAPPING_VOLUME_ID,
        BLOCK_DEVICE_MAPPING_SNAPSHOT_ID,
        BLOCK_DEVICE_MAPPING_VOLUME_SIZE,
        BLOCK_DEVICE_MAPPING_DELETE_ON_TERM,
    ) = (
        'device_name', 'volume_id',
        'snapshot_id',
        'volume_size',
        'delete_on_termination',
    )

    _BLOCK_DEVICE_MAPPING_V2_KEYS = (
        BLOCK_DEVICE_MAPPING_DEVICE_NAME,
        BLOCK_DEVICE_MAPPING_VOLUME_ID,
        BLOCK_DEVICE_MAPPING_IMAGE_ID,
        BLOCK_DEVICE_MAPPING_IMAGE,
        BLOCK_DEVICE_MAPPING_SNAPSHOT_ID,
        BLOCK_DEVICE_MAPPING_SWAP_SIZE,
        BLOCK_DEVICE_MAPPING_DEVICE_TYPE,
        BLOCK_DEVICE_MAPPING_DISK_BUS,
        BLOCK_DEVICE_MAPPING_BOOT_INDEX,
        BLOCK_DEVICE_MAPPING_VOLUME_SIZE,
        BLOCK_DEVICE_MAPPING_DELETE_ON_TERM,
        BLOCK_DEVICE_MAPPING_EPHEMERAL_SIZE,
        BLOCK_DEVICE_MAPPING_EPHEMERAL_FORMAT,
    ) = (
        'device_name',
        'volume_id',
        'image_id',
        'image',
        'snapshot_id',
        'swap_size',
        'device_type',
        'disk_bus',
        'boot_index',
        'volume_size',
        'delete_on_termination',
        'ephemeral_size',
        'ephemeral_format'
    )

    _NETWORK_KEYS = (
        NETWORK_UUID, NETWORK_ID, NETWORK_FIXED_IP, NETWORK_PORT,
        NETWORK_SUBNET, NETWORK_PORT_EXTRA, NETWORK_FLOATING_IP,
        ALLOCATE_NETWORK, NIC_TAG,
    ) = (
        'uuid', 'network', 'fixed_ip', 'port',
        'subnet', 'port_extra_properties', 'floating_ip',
        'allocate_network', 'tag',
    )

    _IFACE_MANAGED_KEYS = (NETWORK_PORT, NETWORK_ID,
                           NETWORK_FIXED_IP, NETWORK_SUBNET)

    _SOFTWARE_CONFIG_FORMATS = (
        HEAT_CFNTOOLS, RAW, SOFTWARE_CONFIG
    ) = (
        'HEAT_CFNTOOLS', 'RAW', 'SOFTWARE_CONFIG'
    )

    _SOFTWARE_CONFIG_TRANSPORTS = (
        POLL_SERVER_CFN, POLL_SERVER_HEAT, POLL_TEMP_URL, ZAQAR_MESSAGE
    ) = (
        'POLL_SERVER_CFN', 'POLL_SERVER_HEAT', 'POLL_TEMP_URL', 'ZAQAR_MESSAGE'
    )

    _ALLOCATE_TYPES = (
        NETWORK_NONE, NETWORK_AUTO,
    ) = (
        'none', 'auto',
    )

    _DEPLOYMENT_SWIFT_DATA_KEYS = (
        CONTAINER, OBJECT
    ) = (
        'container', 'object',
    )

    ATTRIBUTES = (
        NAME_ATTR, ADDRESSES, NETWORKS_ATTR, FIRST_ADDRESS,
        INSTANCE_NAME, ACCESSIPV4, ACCESSIPV6, CONSOLE_URLS, TAGS_ATTR,
        OS_COLLECT_CONFIG
    ) = (
        'name', 'addresses', 'networks', 'first_address',
        'instance_name', 'accessIPv4', 'accessIPv6', 'console_urls', 'tags',
        'os_collect_config'
    )

    # Image Statuses
    IMAGE_STATUSES = (IMAGE_ACTIVE, IMAGE_ERROR,
                      IMAGE_DELETED) = ('active', 'error', 'deleted')

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Server name.'),
            update_allowed=True
        ),
        IMAGE: properties.Schema(
            properties.Schema.STRING,
            _('The ID or name of the image to boot with.'),
            constraints=[
                constraints.CustomConstraint('glance.image')
            ],
            update_allowed=True
        ),
        BLOCK_DEVICE_MAPPING: properties.Schema(
            properties.Schema.LIST,
            _('Block device mappings for this server.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    BLOCK_DEVICE_MAPPING_DEVICE_NAME: properties.Schema(
                        properties.Schema.STRING,
                        _('A device name where the volume will be '
                          'attached in the system at /dev/device_name. '
                          'This value is typically vda.'),
                        required=True
                    ),
                    BLOCK_DEVICE_MAPPING_VOLUME_ID: properties.Schema(
                        properties.Schema.STRING,
                        _('The ID of the volume to boot from. Only one '
                          'of volume_id or snapshot_id should be '
                          'provided.'),
                        constraints=[
                            constraints.CustomConstraint('cinder.volume')
                        ]
                    ),
                    BLOCK_DEVICE_MAPPING_SNAPSHOT_ID: properties.Schema(
                        properties.Schema.STRING,
                        _('The ID of the snapshot to create a volume '
                          'from.'),
                        constraints=[
                            constraints.CustomConstraint('cinder.snapshot')
                        ]
                    ),
                    BLOCK_DEVICE_MAPPING_VOLUME_SIZE: properties.Schema(
                        properties.Schema.INTEGER,
                        _('The size of the volume, in GB. It is safe to '
                          'leave this blank and have the Compute service '
                          'infer the size.')
                    ),
                    BLOCK_DEVICE_MAPPING_DELETE_ON_TERM: properties.Schema(
                        properties.Schema.BOOLEAN,
                        _('Indicate whether the volume should be deleted '
                          'when the server is terminated.')
                    ),
                },
            )
        ),
        BLOCK_DEVICE_MAPPING_V2: properties.Schema(
            properties.Schema.LIST,
            _('Block device mappings v2 for this server.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    BLOCK_DEVICE_MAPPING_DEVICE_NAME: properties.Schema(
                        properties.Schema.STRING,
                        _('A device name where the volume will be '
                          'attached in the system at /dev/device_name. '
                          'This value is typically vda.'),
                    ),
                    BLOCK_DEVICE_MAPPING_VOLUME_ID: properties.Schema(
                        properties.Schema.STRING,
                        _('The volume_id can be boot or non-boot device '
                          'to the server.'),
                        constraints=[
                            constraints.CustomConstraint('cinder.volume')
                        ]
                    ),
                    BLOCK_DEVICE_MAPPING_IMAGE_ID: properties.Schema(
                        properties.Schema.STRING,
                        _('The ID of the image to create a volume from.'),
                        support_status=support.SupportStatus(
                            status=support.HIDDEN,
                            version='9.0.0',
                            message=_('Use property %s.') %
                                    BLOCK_DEVICE_MAPPING_IMAGE,
                            previous_status=support.SupportStatus(
                                status=support.DEPRECATED,
                                version='7.0.0',
                                previous_status=support.SupportStatus(
                                    version='5.0.0')
                            )
                        ),
                        constraints=[
                            constraints.CustomConstraint('glance.image')
                        ],
                    ),
                    BLOCK_DEVICE_MAPPING_IMAGE: properties.Schema(
                        properties.Schema.STRING,
                        _('The ID or name of the image '
                          'to create a volume from.'),
                        support_status=support.SupportStatus(version='7.0.0'),
                        constraints=[
                            constraints.CustomConstraint('glance.image')
                        ],
                    ),
                    BLOCK_DEVICE_MAPPING_SNAPSHOT_ID: properties.Schema(
                        properties.Schema.STRING,
                        _('The ID of the snapshot to create a volume '
                          'from.'),
                        constraints=[
                            constraints.CustomConstraint('cinder.snapshot')
                        ]
                    ),
                    BLOCK_DEVICE_MAPPING_SWAP_SIZE: properties.Schema(
                        properties.Schema.INTEGER,
                        _('The size of the swap, in MB.')
                    ),
                    BLOCK_DEVICE_MAPPING_EPHEMERAL_SIZE: properties.Schema(
                        properties.Schema.INTEGER,
                        _('The size of the local ephemeral block device, '
                          'in GB.'),
                        support_status=support.SupportStatus(version='8.0.0'),
                        constraints=[constraints.Range(min=1)]
                    ),
                    BLOCK_DEVICE_MAPPING_EPHEMERAL_FORMAT: properties.Schema(
                        properties.Schema.STRING,
                        _('The format of the local ephemeral block device. '
                          'If no format is specified, uses default value, '
                          'defined in nova configuration file.'),
                        constraints=[
                            constraints.AllowedValues(['ext2', 'ext3', 'ext4',
                                                       'xfs', 'ntfs'])
                        ],
                        support_status=support.SupportStatus(version='8.0.0')
                    ),
                    BLOCK_DEVICE_MAPPING_DEVICE_TYPE: properties.Schema(
                        properties.Schema.STRING,
                        _('Device type: at the moment we can make distinction '
                          'only between disk and cdrom.'),
                        constraints=[
                            constraints.AllowedValues(['cdrom', 'disk']),
                        ],
                    ),
                    BLOCK_DEVICE_MAPPING_DISK_BUS: properties.Schema(
                        properties.Schema.STRING,
                        _('Bus of the device: hypervisor driver chooses a '
                          'suitable default if omitted.'),
                        constraints=[
                            constraints.AllowedValues(['ide', 'lame_bus',
                                                       'scsi', 'usb',
                                                       'virtio']),
                        ],
                    ),
                    BLOCK_DEVICE_MAPPING_BOOT_INDEX: properties.Schema(
                        properties.Schema.INTEGER,
                        _('Integer used for ordering the boot disks. If '
                          'it is not specified, value "0" will be set '
                          'for bootable sources (volume, snapshot, image); '
                          'value "-1" will be set for non-bootable sources.'),
                    ),
                    BLOCK_DEVICE_MAPPING_VOLUME_SIZE: properties.Schema(
                        properties.Schema.INTEGER,
                        _('Size of the block device in GB. If it is omitted, '
                          'hypervisor driver calculates size.'),
                    ),
                    BLOCK_DEVICE_MAPPING_DELETE_ON_TERM: properties.Schema(
                        properties.Schema.BOOLEAN,
                        _('Indicate whether the volume should be deleted '
                          'when the server is terminated. '
                          'Defaults to "False" in case of a volume, snapshot '
                          'or image and to "True" in case of swap or '
                          'ephemeral.')
                    ),
                },
            ),
            support_status=support.SupportStatus(version='2015.1')
        ),
        FLAVOR: properties.Schema(
            properties.Schema.STRING,
            _('The ID or name of the flavor to boot onto.'),
            required=True,
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('nova.flavor')
            ]
        ),
        FLAVOR_UPDATE_POLICY: properties.Schema(
            properties.Schema.STRING,
            _('Policy on how to apply a flavor update; either by requesting '
              'a server resize or by replacing the entire server.'),
            default='RESIZE',
            constraints=[
                constraints.AllowedValues(['RESIZE', 'REPLACE']),
            ],
            update_allowed=True
        ),
        IMAGE_UPDATE_POLICY: properties.Schema(
            properties.Schema.STRING,
            _('Policy on how to apply an image-id update; either by '
              'requesting a server rebuild or by replacing '
              'the entire server.'),
            default='REBUILD',
            constraints=[
                constraints.AllowedValues(['REBUILD', 'REPLACE',
                                           'REBUILD_PRESERVE_EPHEMERAL']),
            ],
            update_allowed=True
        ),
        KEY_NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of keypair to inject into the server.'),
            constraints=[
                constraints.CustomConstraint('nova.keypair')
            ]
        ),
        ADMIN_USER: properties.Schema(
            properties.Schema.STRING,
            _('Name of the administrative user to use on the server.'),
            support_status=support.SupportStatus(
                status=support.HIDDEN,
                version='5.0.0',
                message=_('The default cloud-init user set up for each image '
                          '(e.g. "ubuntu" for Ubuntu 12.04+, "fedora" for '
                          'Fedora 19+ and "cloud-user" for CentOS/RHEL 6.5).'),
                previous_status=support.SupportStatus(
                    status=support.DEPRECATED,
                    version='2014.1',
                    previous_status=support.SupportStatus(version='2013.2')
                )
            )
        ),
        AVAILABILITY_ZONE: properties.Schema(
            properties.Schema.STRING,
            _('Name of the availability zone for server placement.')
        ),
        SECURITY_GROUPS: properties.Schema(
            properties.Schema.LIST,
            _('List of security group names or IDs. Cannot be used if '
              'neutron ports are associated with this server; assign '
              'security groups to the ports instead.'),
            default=[]
        ),
        NETWORKS: properties.Schema(
            properties.Schema.LIST,
            _('An ordered list of nics to be added to this server, with '
              'information about connected networks, fixed ips, port etc.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    NETWORK_UUID: properties.Schema(
                        properties.Schema.STRING,
                        _('ID of network to create a port on.'),
                        support_status=support.SupportStatus(
                            status=support.HIDDEN,
                            version='5.0.0',
                            previous_status=support.SupportStatus(
                                status=support.DEPRECATED,
                                message=_('Use property %s.') % NETWORK_ID,
                                version='2014.1'
                            )
                        ),
                        constraints=[
                            constraints.CustomConstraint('neutron.network')
                        ]
                    ),
                    NETWORK_ID: properties.Schema(
                        properties.Schema.STRING,
                        _('Name or ID of network to create a port on.'),
                        constraints=[
                            constraints.CustomConstraint('neutron.network')
                        ]
                    ),
                    ALLOCATE_NETWORK: properties.Schema(
                        properties.Schema.STRING,
                        _('The special string values of network, '
                          'auto: means either a network that is already '
                          'available to the project will be used, or if one '
                          'does not exist, will be automatically created for '
                          'the project; none: means no networking will be '
                          'allocated for the created server. Supported by '
                          'Nova API since version "2.37". This property can '
                          'not be used with other network keys.'),
                        support_status=support.SupportStatus(version='9.0.0'),
                        constraints=[
                            constraints.AllowedValues(
                                [NETWORK_NONE, NETWORK_AUTO])
                        ],
                        update_allowed=True,
                    ),
                    NETWORK_FIXED_IP: properties.Schema(
                        properties.Schema.STRING,
                        _('Fixed IP address to specify for the port '
                          'created on the requested network.'),
                        constraints=[
                            constraints.CustomConstraint('ip_addr')
                        ]
                    ),
                    NETWORK_PORT: properties.Schema(
                        properties.Schema.STRING,
                        _('ID of an existing port to associate with this '
                          'server.'),
                        constraints=[
                            constraints.CustomConstraint('neutron.port')
                        ]
                    ),
                    NETWORK_PORT_EXTRA: properties.Schema(
                        properties.Schema.MAP,
                        _('Dict, which has expand properties for port. '
                          'Used only if port property is not specified '
                          'for creating port.'),
                        schema=neutron_port.Port.extra_properties_schema,
                        support_status=support.SupportStatus(version='6.0.0')
                    ),
                    NETWORK_SUBNET: properties.Schema(
                        properties.Schema.STRING,
                        _('Subnet in which to allocate the IP address for '
                          'port. Used for creating port, based on derived '
                          'properties. If subnet is specified, network '
                          'property becomes optional.'),
                        support_status=support.SupportStatus(version='5.0.0')
                    ),
                    NETWORK_FLOATING_IP: properties.Schema(
                        properties.Schema.STRING,
                        _('ID of the floating IP to associate.'),
                        support_status=support.SupportStatus(version='6.0.0')
                    ),
                    NIC_TAG: properties.Schema(
                        properties.Schema.STRING,
                        _('Port tag. Heat ignores any update on this property '
                          'as nova does not support it.'),
                        support_status=support.SupportStatus(version='9.0.0')
                    )
                },
            ),
            update_allowed=True
        ),
        SCHEDULER_HINTS: properties.Schema(
            properties.Schema.MAP,
            _('Arbitrary key-value pairs specified by the client to help '
              'boot a server.')
        ),
        METADATA: properties.Schema(
            properties.Schema.MAP,
            _('Arbitrary key/value metadata to store for this server. Both '
              'keys and values must be 255 characters or less. Non-string '
              'values will be serialized to JSON (and the serialized '
              'string must be 255 characters or less).'),
            update_allowed=True,
            default={}
        ),
        USER_DATA_FORMAT: properties.Schema(
            properties.Schema.STRING,
            _('How the user_data should be formatted for the server. For '
              'HEAT_CFNTOOLS, the user_data is bundled as part of the '
              'heat-cfntools cloud-init boot configuration data. For RAW '
              'the user_data is passed to Nova unmodified. '
              'For SOFTWARE_CONFIG user_data is bundled as part of the '
              'software config data, and metadata is derived from any '
              'associated SoftwareDeployment resources. And if the '
              'user_data is in CoreOS ignition(json) format, the metadata '
              'will be injected into the user_data automatically by Heat.'),
            default=cfg.CONF.default_user_data_format,
            constraints=[
                constraints.AllowedValues(_SOFTWARE_CONFIG_FORMATS),
            ]
        ),
        SOFTWARE_CONFIG_TRANSPORT: properties.Schema(
            properties.Schema.STRING,
            _('How the server should receive the metadata required for '
              'software configuration. POLL_SERVER_CFN will allow calls to '
              'the cfn API action DescribeStackResource authenticated with '
              'the provided keypair. POLL_SERVER_HEAT will allow calls to '
              'the Heat API resource-show using the provided keystone '
              'credentials. POLL_TEMP_URL will create and populate a '
              'Swift TempURL with metadata for polling. ZAQAR_MESSAGE will '
              'create a dedicated zaqar queue and post the metadata '
              'for polling.'),
            default=cfg.CONF.default_software_config_transport,
            update_allowed=True,
            constraints=[
                constraints.AllowedValues(_SOFTWARE_CONFIG_TRANSPORTS),
            ]
        ),
        USER_DATA_UPDATE_POLICY: properties.Schema(
            properties.Schema.STRING,
            _('Policy on how to apply a user_data update; by '
              'ignoring it, by replacing the entire server, '
              'or rebuild the server.'),
            default='REPLACE',
            constraints=[
                constraints.AllowedValues(['REPLACE', 'IGNORE', 'REBUILD']),
            ],
            support_status=support.SupportStatus(version='6.0.0'),
            update_allowed=True
        ),
        USER_DATA: properties.Schema(
            properties.Schema.STRING,
            _('User data script to be executed by cloud-init or CoreOS '
              'ignition. Changes cause replacement of the resource '
              'by default, but can be ignored altogether by setting the '
              '`user_data_update_policy` property.'),
            default='',
            update_allowed=True
        ),
        RESERVATION_ID: properties.Schema(
            properties.Schema.STRING,
            _('A UUID for the set of servers being requested.')
        ),
        CONFIG_DRIVE: properties.Schema(
            properties.Schema.BOOLEAN,
            _('If True, enable config drive on the server.')
        ),
        DISK_CONFIG: properties.Schema(
            properties.Schema.STRING,
            _('Control how the disk is partitioned when the server is '
              'created.'),
            constraints=[
                constraints.AllowedValues(['AUTO', 'MANUAL']),
            ]
        ),
        PERSONALITY: properties.Schema(
            properties.Schema.MAP,
            _('A map of files to create/overwrite on the server upon boot. '
              'Keys are file names and values are the file contents.'),
            support_status=support.SupportStatus(
                status=support.DEPRECATED,
                version='12.0.0',
                message=_('This is not supported with nova api '
                          'microversion 2.57 and above. '
                          'OS::Nova::Server resource will not support '
                          'it in the future. Please use user_data or metadata '
                          'instead. However, you can set heat config option '
                          'max_nova_api_microversion < 2.57 to use '
                          'this property in the meantime.')
            ),
            default={}
        ),
        ADMIN_PASS: properties.Schema(
            properties.Schema.STRING,
            _('The administrator password for the server.'),
            update_allowed=True
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            _('Server tags. Supported since client version 2.26.'),
            support_status=support.SupportStatus(version='8.0.0'),
            schema=properties.Schema(properties.Schema.STRING),
            update_allowed=True
        ),
        DEPLOYMENT_SWIFT_DATA: properties.Schema(
            properties.Schema.MAP,
            _('Swift container and object to use for storing deployment data '
              'for the server resource. The parameter is a map value '
              'with the keys "container" and "object", and the values '
              'are the corresponding container and object names. The '
              'software_config_transport parameter must be set to '
              'POLL_TEMP_URL for swift to be used. If not specified, '
              'and software_config_transport is set to POLL_TEMP_URL, a '
              'container will be automatically created from the resource '
              'name, and the object name will be a generated uuid.'),
            support_status=support.SupportStatus(version='9.0.0'),
            default={},
            update_allowed=True,
            schema={
                CONTAINER: properties.Schema(
                    properties.Schema.STRING,
                    _('Name of the container.'),
                    constraints=[
                        constraints.Length(min=1)
                    ]
                ),
                OBJECT: properties.Schema(
                    properties.Schema.STRING,
                    _('Name of the object.'),
                    constraints=[
                        constraints.Length(min=1)
                    ]
                )
            }
        )
    }

    attributes_schema = {
        NAME_ATTR: attributes.Schema(
            _('Name of the server.'),
            type=attributes.Schema.STRING
        ),
        ADDRESSES: attributes.Schema(
            _('A dict of all network addresses with corresponding port_id and '
              'subnets. Each network will have two keys in dict, they are '
              'network name and network id. The port ID may be obtained '
              'through the following expression: ``{get_attr: [<server>, '
              'addresses, <network name_or_id>, 0, port]}``. The subnets may '
              'be obtained trough the following expression: ``{get_attr: '
              '[<server>, addresses, <network name_or_id>, 0, subnets]}``. '
              'The network may be obtained through the following expression: '
              '``{get_attr: [<server>, addresses, <network name_or_id>, 0, '
              'network]}``.'),
            type=attributes.Schema.MAP,
            support_status=support.SupportStatus(
                version='11.0.0',
                status=support.SUPPORTED,
                message=_('The attribute was extended to include subnets and '
                          'network with version 11.0.0.'),
                previous_status=support.SupportStatus(
                    status=support.SUPPORTED
                )
            )
        ),
        NETWORKS_ATTR: attributes.Schema(
            _('A dict of assigned network addresses of the form: '
              '{"public": [ip1, ip2...], "private": [ip3, ip4], '
              '"public_uuid": [ip1, ip2...], "private_uuid": [ip3, ip4]}. '
              'Each network will have two keys in dict, they are network '
              'name and network id.'),
            type=attributes.Schema.MAP
        ),
        FIRST_ADDRESS: attributes.Schema(
            _('Convenience attribute to fetch the first assigned network '
              'address, or an empty string if nothing has been assigned at '
              'this time. Result may not be predictable if the server has '
              'addresses from more than one network.'),
            support_status=support.SupportStatus(
                status=support.HIDDEN,
                version='5.0.0',
                message=_('Use the networks attribute instead of '
                          'first_address. For example: "{get_attr: '
                          '[<server name>, networks, <network name>, 0]}"'),
                previous_status=support.SupportStatus(
                    status=support.DEPRECATED,
                    version='2014.2',
                    previous_status=support.SupportStatus(version='2013.2')
                )
            )
        ),
        INSTANCE_NAME: attributes.Schema(
            _('AWS compatible instance name.'),
            type=attributes.Schema.STRING
        ),
        ACCESSIPV4: attributes.Schema(
            _('The manually assigned alternative public IPv4 address '
              'of the server.'),
            type=attributes.Schema.STRING,
            support_status=support.SupportStatus(
                status=support.DEPRECATED,
                version='14.0.0',
                previous_status=support.SupportStatus(version='2015.1')),
        ),
        ACCESSIPV6: attributes.Schema(
            _('The manually assigned alternative public IPv6 address '
              'of the server.'),
            type=attributes.Schema.STRING,
            support_status=support.SupportStatus(
                status=support.DEPRECATED,
                version='14.0.0',
                previous_status=support.SupportStatus(version='2015.1'))
        ),
        CONSOLE_URLS: attributes.Schema(
            _("URLs of server's consoles. "
              "To get a specific console type, the requested type "
              "can be specified as parameter to the get_attr function, "
              "e.g. get_attr: [ <server>, console_urls, novnc ]. "
              "Currently supported types are "
              "novnc, xvpvnc, spice-html5, rdp-html5, serial and webmks."),
            support_status=support.SupportStatus(version='2015.1'),
            type=attributes.Schema.MAP
        ),
        TAGS_ATTR: attributes.Schema(
            _('Tags from the server. Supported since client version 2.26.'),
            support_status=support.SupportStatus(version='8.0.0'),
            type=attributes.Schema.LIST
        ),
        OS_COLLECT_CONFIG: attributes.Schema(
            _('The os-collect-config configuration for the server\'s local '
              'agent to be configured to connect to Heat to retrieve '
              'deployment data.'),
            support_status=support.SupportStatus(version='9.0.0'),
            type=attributes.Schema.MAP,
            cache_mode=attributes.Schema.CACHE_NONE
        ),
    }

    default_client_name = 'nova'

    def translation_rules(self, props):
        neutron_client_plugin = self.client_plugin('neutron')
        glance_client_plugin = self.client_plugin('glance')
        rules = [
            translation.TranslationRule(
                props,
                translation.TranslationRule.REPLACE,
                translation_path=[self.NETWORKS, self.NETWORK_ID],
                value_name=self.NETWORK_UUID),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.FLAVOR],
                client_plugin=self.client_plugin('nova'),
                finder='find_flavor_by_name_or_id'),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.IMAGE],
                client_plugin=glance_client_plugin,
                finder='find_image_by_name_or_id'),
            translation.TranslationRule(
                props,
                translation.TranslationRule.REPLACE,
                translation_path=[self.BLOCK_DEVICE_MAPPING_V2,
                                  self.BLOCK_DEVICE_MAPPING_IMAGE],
                value_name=self.BLOCK_DEVICE_MAPPING_IMAGE_ID),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.BLOCK_DEVICE_MAPPING_V2,
                                  self.BLOCK_DEVICE_MAPPING_IMAGE],
                client_plugin=glance_client_plugin,
                finder='find_image_by_name_or_id'),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.NETWORKS, self.NETWORK_ID],
                client_plugin=neutron_client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=neutron_client_plugin.RES_TYPE_NETWORK),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.NETWORKS, self.NETWORK_SUBNET],
                client_plugin=neutron_client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=neutron_client_plugin.RES_TYPE_SUBNET),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.NETWORKS, self.NETWORK_PORT],
                client_plugin=neutron_client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=neutron_client_plugin.RES_TYPE_PORT)
        ]
        return rules

    def __init__(self, name, json_snippet, stack):
        super(Server, self).__init__(name, json_snippet, stack)
        if self.user_data_software_config():
            self._register_access_key()
        self.default_collectors = ['ec2']

    def _config_drive(self):
        # This method is overridden by the derived CloudServer resource
        return self.properties[self.CONFIG_DRIVE]

    def user_data_raw(self):
        return self.properties[self.USER_DATA_FORMAT] == self.RAW

    def user_data_software_config(self):
        return self.properties[
            self.USER_DATA_FORMAT] == self.SOFTWARE_CONFIG

    def get_software_config(self, ud_content):
        with self.rpc_client().ignore_error_by_name('NotFound'):
            sc = self.rpc_client().show_software_config(
                self.context, ud_content)
            return sc[rpc_api.SOFTWARE_CONFIG_CONFIG]
        return ud_content

    def handle_create(self):
        security_groups = self.properties[self.SECURITY_GROUPS]

        user_data_format = self.properties[self.USER_DATA_FORMAT]
        ud_content = self.properties[self.USER_DATA]
        if self.user_data_software_config() or self.user_data_raw():
            if uuidutils.is_uuid_like(ud_content):
                # attempt to load the userdata from software config
                ud_content = self.get_software_config(ud_content)

        metadata = self.metadata_get(True) or {}

        if self.user_data_software_config():
            self._create_transport_credentials(self.properties)
            self._populate_deployments_metadata(metadata, self.properties)

        userdata = self.client_plugin().build_userdata(
            metadata,
            ud_content,
            instance_user=None,
            user_data_format=user_data_format)

        availability_zone = self.properties[self.AVAILABILITY_ZONE]
        instance_meta = self.properties[self.METADATA]
        if instance_meta:
            instance_meta = self.client_plugin().meta_serialize(
                instance_meta)

        scheduler_hints = self._scheduler_hints(
            self.properties[self.SCHEDULER_HINTS])

        nics = self._build_nics(self.properties[self.NETWORKS],
                                security_groups=security_groups)
        block_device_mapping = self._build_block_device_mapping(
            self.properties[self.BLOCK_DEVICE_MAPPING])
        block_device_mapping_v2 = self._build_block_device_mapping_v2(
            self.properties[self.BLOCK_DEVICE_MAPPING_V2])
        reservation_id = self.properties[self.RESERVATION_ID]
        disk_config = self.properties[self.DISK_CONFIG]
        admin_pass = self.properties[self.ADMIN_PASS] or None
        personality_files = self.properties[self.PERSONALITY]
        key_name = self.properties[self.KEY_NAME]
        flavor = self.properties[self.FLAVOR]
        image = self.properties[self.IMAGE]

        server = None
        try:
            server = self.client().servers.create(
                name=self._server_name(),
                image=image,
                flavor=flavor,
                key_name=key_name,
                security_groups=security_groups,
                userdata=userdata,
                meta=instance_meta,
                scheduler_hints=scheduler_hints,
                nics=nics,
                availability_zone=availability_zone,
                block_device_mapping=block_device_mapping,
                block_device_mapping_v2=block_device_mapping_v2,
                reservation_id=reservation_id,
                config_drive=self._config_drive(),
                disk_config=disk_config,
                files=personality_files,
                admin_pass=admin_pass)
        finally:
            # Avoid a race condition where the thread could be canceled
            # before the ID is stored
            if server is not None:
                self.resource_id_set(server.id)

        assert server is not None
        return server.id

    def check_create_complete(self, server_id):
        check = self.client_plugin()._check_active(server_id)
        if check:
            if self.properties[self.TAGS]:
                self._update_server_tags(self.properties[self.TAGS])
            self.store_external_ports()
        return check

    def _update_server_tags(self, tags):
        server = self.client().servers.get(self.resource_id)
        self.client().servers.set_tags(server, tags)

    def handle_check(self):
        server = self.client().servers.get(self.resource_id)
        status = self.client_plugin().get_status(server)
        checks = [{'attr': 'status', 'expected': 'ACTIVE', 'current': status}]
        self._verify_check_conditions(checks)

    def get_live_resource_data(self):
        try:
            server = self.client().servers.get(self.resource_id)
            server_data = server.to_dict()
            active = self.client_plugin()._check_active(server)
            if not active:
                # There is no difference what error raised, because update
                # method of resource just silently log it as warning.
                raise exception.Error(_('Server %s is not '
                                        'in ACTIVE state') % self.name)
        except Exception as ex:
            if self.client_plugin().is_not_found(ex):
                raise exception.EntityNotFound(entity='Resource',
                                               name=self.name)
            raise

        if self.client_plugin().is_version_supported(MICROVERSION_TAGS):
            tag_server = self.client().servers.get(self.resource_id)
            server_data['tags'] = tag_server.tag_list()
        return server, server_data

    def parse_live_resource_data(self, resource_properties, resource_data):
        server, server_data = resource_data
        flavor = server_data.get(self.FLAVOR)
        # NOTE(pas-ha) since compute API 2.47 flavor in instance
        # does not have "id" but "original_name" instead,
        # check for both here, and fail if none of them are in flavor.
        if "id" in flavor:
            flavor_value = flavor["id"]
        elif "original_name" in flavor:
            flavor_value = flavor["original_name"]
        else:
            raise KeyError("Flavor does not contain id or original_name")
        result = {
            # there's a risk that flavor id will be int type, so cast to str
            self.FLAVOR: str(flavor_value),
            self.IMAGE: str(server_data.get(self.IMAGE)['id']),
            self.NAME: server_data.get(self.NAME),
            self.METADATA: server_data.get(self.METADATA),
            self.NETWORKS: self._get_live_networks(server, resource_properties)
        }
        if 'tags' in server_data:
            result.update({self.TAGS: server_data['tags']})
        return result

    def _get_live_networks(self, server, props):
        reality_nets = self._get_server_addresses(server,
                                                  extend_networks=False)
        reality_net_ids = {}
        client_plugin = self.client_plugin('neutron')
        for net_key in reality_nets:
            try:
                net_id = client_plugin.find_resourceid_by_name_or_id(
                    client_plugin.RES_TYPE_NETWORK,
                    net_key)
            except Exception as ex:
                if (client_plugin.is_not_found(ex) or
                        client_plugin.is_no_unique(ex)):
                    net_id = None
                else:
                    raise
            if net_id:
                reality_net_ids[net_id] = reality_nets.get(net_key)

        resource_nets = props.get(self.NETWORKS)

        result_nets = []
        for net in resource_nets or []:
            net_id = self._get_network_id(net)
            if reality_net_ids.get(net_id):
                for idx, address in enumerate(reality_net_ids.get(net_id)):
                    if address['addr'] == net[self.NETWORK_FIXED_IP]:
                        result_nets.append(net)
                        reality_net_ids.get(net_id).pop(idx)
                        break

        for key, value in reality_nets.items():
            for address in reality_nets[key]:
                new_net = {self.NETWORK_ID: key,
                           self.NETWORK_FIXED_IP: address['addr']}
                if address['port'] not in [port['id']
                                           for port in self._data_get_ports()]:
                    new_net.update({self.NETWORK_PORT: address['port']})
                result_nets.append(new_net)
        return result_nets

    @classmethod
    def _build_block_device_mapping(cls, bdm):
        if not bdm:
            return None
        bdm_dict = {}
        for mapping in bdm:
            mapping_parts = []
            snapshot_id = mapping.get(cls.BLOCK_DEVICE_MAPPING_SNAPSHOT_ID)
            if snapshot_id:
                mapping_parts.append(snapshot_id)
                mapping_parts.append('snap')
            else:
                volume_id = mapping.get(cls.BLOCK_DEVICE_MAPPING_VOLUME_ID)
                mapping_parts.append(volume_id)
                mapping_parts.append('')

            volume_size = mapping.get(cls.BLOCK_DEVICE_MAPPING_VOLUME_SIZE)
            delete = mapping.get(cls.BLOCK_DEVICE_MAPPING_DELETE_ON_TERM)
            if volume_size:
                mapping_parts.append(str(volume_size))
            else:
                mapping_parts.append('')
            if delete:
                mapping_parts.append(str(delete))

            device_name = mapping.get(cls.BLOCK_DEVICE_MAPPING_DEVICE_NAME)
            bdm_dict[device_name] = ':'.join(mapping_parts)

        return bdm_dict

    @classmethod
    def _build_block_device_mapping_v2(cls, bdm_v2):
        if not bdm_v2:
            return None

        bdm_v2_list = []
        for mapping in bdm_v2:
            bmd_dict = None
            if mapping.get(cls.BLOCK_DEVICE_MAPPING_VOLUME_ID):
                bmd_dict = {
                    'uuid': mapping.get(cls.BLOCK_DEVICE_MAPPING_VOLUME_ID),
                    'source_type': 'volume',
                    'destination_type': 'volume',
                    'boot_index': 0,
                    'delete_on_termination': False,
                }
            elif mapping.get(cls.BLOCK_DEVICE_MAPPING_SNAPSHOT_ID):
                bmd_dict = {
                    'uuid': mapping.get(cls.BLOCK_DEVICE_MAPPING_SNAPSHOT_ID),
                    'source_type': 'snapshot',
                    'destination_type': 'volume',
                    'boot_index': 0,
                    'delete_on_termination': False,
                }
            elif mapping.get(cls.BLOCK_DEVICE_MAPPING_IMAGE):
                bmd_dict = {
                    'uuid': mapping.get(cls.BLOCK_DEVICE_MAPPING_IMAGE),
                    'source_type': 'image',
                    'destination_type': 'volume',
                    'boot_index': 0,
                    'delete_on_termination': False,
                }
            elif mapping.get(cls.BLOCK_DEVICE_MAPPING_SWAP_SIZE):
                bmd_dict = {
                    'source_type': 'blank',
                    'destination_type': 'local',
                    'boot_index': -1,
                    'delete_on_termination': True,
                    'guest_format': 'swap',
                    'volume_size': mapping.get(
                        cls.BLOCK_DEVICE_MAPPING_SWAP_SIZE),
                }
            elif (mapping.get(cls.BLOCK_DEVICE_MAPPING_EPHEMERAL_SIZE) or
                  mapping.get(cls.BLOCK_DEVICE_MAPPING_EPHEMERAL_FORMAT)):
                bmd_dict = {
                    'source_type': 'blank',
                    'destination_type': 'local',
                    'boot_index': -1,
                    'delete_on_termination': True
                }
                ephemeral_size = mapping.get(
                    cls.BLOCK_DEVICE_MAPPING_EPHEMERAL_SIZE)
                if ephemeral_size:
                    bmd_dict.update({'volume_size': ephemeral_size})
                ephemeral_format = mapping.get(
                    cls.BLOCK_DEVICE_MAPPING_EPHEMERAL_FORMAT)
                if ephemeral_format:
                    bmd_dict.update({'guest_format': ephemeral_format})

            # NOTE(prazumovsky): In case of server doesn't take empty value of
            # device name, need to escape from such situation.
            device_name = mapping.get(cls.BLOCK_DEVICE_MAPPING_DEVICE_NAME)
            if device_name:
                bmd_dict[cls.BLOCK_DEVICE_MAPPING_DEVICE_NAME] = device_name

            update_props = (cls.BLOCK_DEVICE_MAPPING_DEVICE_TYPE,
                            cls.BLOCK_DEVICE_MAPPING_DISK_BUS,
                            cls.BLOCK_DEVICE_MAPPING_BOOT_INDEX,
                            cls.BLOCK_DEVICE_MAPPING_VOLUME_SIZE,
                            cls.BLOCK_DEVICE_MAPPING_DELETE_ON_TERM)

            for update_prop in update_props:
                if mapping.get(update_prop) is not None:
                    bmd_dict[update_prop] = mapping.get(update_prop)

            if bmd_dict:
                bdm_v2_list.append(bmd_dict)

        return bdm_v2_list

    def _get_subnets_attr(self, fixed_ips):
        subnets = []
        try:
            for fixed_ip in fixed_ips:
                if fixed_ip.get('subnet_id'):
                    subnets.append(self.client('neutron').show_subnet(
                        fixed_ip['subnet_id'])['subnet'])
        except Exception as ex:
            LOG.warning("Failed to fetch resource attributes: %s", ex)
            return
        return subnets

    def _get_network_attr(self, network_id):
        try:
            return self.client('neutron').show_network(network_id)['network']
        except Exception as ex:
            LOG.warning("Failed to fetch resource attributes: %s", ex)
            return

    def _get_server_addresses(self, server, extend_networks=True):
        """Adds port id, subnets and network attributes to addresses list.

        This method is used only for resolving attributes.
        :param server: The server resource
        :param extend_networks: When False the network is not extended, i.e
                                the net is returned without replacing name on
                                id.
        """
        nets = {}
        ifaces = self.client('neutron').list_ports(device_id=server.id)
        for port in ifaces['ports']:
            net_label = self.client('neutron').list_networks(
                id=port['network_id'])['networks'][0]['name']
            net = nets.setdefault(net_label, [])
            for fixed_ip in port['fixed_ips']:
                addr = {'addr': fixed_ip.get('ip_address'),
                        'OS-EXT-IPS-MAC:mac_addr': port['mac_address'],
                        'OS-EXT-IPS:type': 'fixed',
                        'port': port['id']}

                try:
                    addr['version'] = ipaddress.ip_address(
                        addr['addr']).version,
                except ValueError:
                    addr['version'] = None

                if addr['addr']:
                    fips = self.client('neutron').list_floatingips(
                        fixed_ip_address=addr['addr'])
                    for fip in fips['floatingips']:
                        net.append({
                            'addr': fip['floating_ip_address'],
                            'version': addr['version'],
                            'OS-EXT-IPS-MAC:mac_addr': port['mac_address'],
                            'OS-EXT-IPS:type': 'floating',
                            'port': None})

                # _get_live_networks() uses this method to get reality_nets.
                # We don't need to get subnets and network in that case. Only
                # do the external calls if extend_networks is true, i.e called
                # from _resolve_attribute()
                if not extend_networks:
                    net.append(addr)
                    continue

                addr['subnets'] = self._get_subnets_attr(port['fixed_ips'])
                addr['network'] = self._get_network_attr(port['network_id'])

                net.append(addr)

        if extend_networks:
            return self._extend_networks(nets)
        else:
            return nets

    def _extend_networks(self, networks):
        """Method adds same networks with replaced name on network id.

        This method is used only for resolving attributes.
        """
        nets = copy.deepcopy(networks)
        client_plugin = self.client_plugin('neutron')
        for key in list(nets.keys()):
            try:
                net_id = client_plugin.find_resourceid_by_name_or_id(
                    client_plugin.RES_TYPE_NETWORK,
                    key)
            except Exception as ex:
                if (client_plugin.is_not_found(ex) or
                        client_plugin.is_no_unique(ex)):
                    net_id = None
                else:
                    raise
            if net_id:
                nets[net_id] = nets[key]
        return nets

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        if name == self.FIRST_ADDRESS:
            return self.client_plugin().server_to_ipaddress(
                self.resource_id) or ''
        if name == self.OS_COLLECT_CONFIG:
            return self.metadata_get().get('os-collect-config', {})
        if name == self.NAME_ATTR:
            return self._server_name()
        try:
            server = self.client().servers.get(self.resource_id)
        except Exception as e:
            self.client_plugin().ignore_not_found(e)
            return ''
        if name == self.ADDRESSES:
            return self._get_server_addresses(server)
        if name == self.NETWORKS_ATTR:
            return self._extend_networks(server.networks)
        if name == self.INSTANCE_NAME:
            return getattr(server, 'OS-EXT-SRV-ATTR:instance_name', None)
        if name == self.ACCESSIPV4:
            return server.accessIPv4
        if name == self.ACCESSIPV6:
            return server.accessIPv6
        if name == self.CONSOLE_URLS:
            return self.client_plugin('nova').get_console_urls(server)
        if name == self.TAGS_ATTR:
            if self.client_plugin().is_version_supported(MICROVERSION_TAGS):
                return self.client().servers.tag_list(server)
            return None

    def add_dependencies(self, deps):
        super(Server, self).add_dependencies(deps)
        # Depend on any Subnet in this template with the same
        # network_id as the networks attached to this server.
        # It is not known which subnet a server might be assigned
        # to so all subnets in a network should be created before
        # the servers in that network.
        try:
            nets = self.properties[self.NETWORKS]
        except (ValueError, TypeError):
            # Properties errors will be caught later in validation,
            # where we can report them in their proper context.
            return
        if not nets:
            return
        for res in self.stack.values():
            if res.has_interface('OS::Neutron::Subnet'):
                try:
                    subnet_net = res.properties.get(subnet.Subnet.NETWORK)
                except (ValueError, TypeError):
                    # Properties errors will be caught later in validation,
                    # where we can report them in their proper context.
                    continue
                # Be wary of the case where we do not know a subnet's
                # network. If that's the case, be safe and add it as a
                # dependency.
                if not subnet_net:
                    deps += (self, res)
                    continue
                for net in nets:
                    # worry about network_id because that could be the match
                    # assigned to the subnet as well and could have been
                    # created by this stack. Regardless, the server should
                    # still wait on the subnet.
                    net_id = net.get(self.NETWORK_ID)
                    if net_id and net_id == subnet_net:
                        deps += (self, res)
                        break
                    # If we don't know a given net_id right now, it's
                    # plausible this subnet depends on it.
                    if not net_id:
                        deps += (self, res)
                        break

    def _update_flavor(self, after_props):
        flavor = after_props[self.FLAVOR]
        handler_args = checker_args = {'args': (flavor,)}
        prg_resize = progress.ServerUpdateProgress(self.resource_id,
                                                   'resize',
                                                   handler_extra=handler_args,
                                                   checker_extra=checker_args)
        prg_verify = progress.ServerUpdateProgress(self.resource_id,
                                                   'verify_resize')
        return prg_resize, prg_verify

    def _update_image(self, after_props):
        image_update_policy = after_props[self.IMAGE_UPDATE_POLICY]

        instance_meta = after_props[self.METADATA]
        if instance_meta is not None:
            instance_meta = self.client_plugin().meta_serialize(
                instance_meta)
        personality_files = after_props[self.PERSONALITY]

        image = after_props[self.IMAGE]
        preserve_ephemeral = (
            image_update_policy == 'REBUILD_PRESERVE_EPHEMERAL')
        password = after_props[self.ADMIN_PASS]
        kwargs = {'password': password,
                  'preserve_ephemeral': preserve_ephemeral,
                  'meta': instance_meta,
                  'files': personality_files}
        prg = progress.ServerUpdateProgress(self.resource_id,
                                            'rebuild',
                                            handler_extra={'args': (image,),
                                                           'kwargs': kwargs})
        return prg

    def _update_user_data_rebuild(self, after_props):
        user_data = after_props[self.USER_DATA]
        prg = progress.ServerUpdateProgress(
            self.resource_id,
            'rebuild',
            handler_extra={'args': (user_data,)})
        return prg

    def _update_networks(self, server, after_props):
        updaters = []
        new_networks = after_props[self.NETWORKS]
        old_networks = self.properties[self.NETWORKS]
        security_groups = after_props[self.SECURITY_GROUPS]

        if not server:
            server = self.client().servers.get(self.resource_id)
        interfaces = server.interface_list()
        remove_ports, add_nets = self.calculate_networks(
            old_networks, new_networks, interfaces, security_groups)

        for port in remove_ports:
            updaters.append(
                progress.ServerUpdateProgress(
                    self.resource_id, 'interface_detach',
                    handler_extra={'args': (port,)},
                    checker_extra={'args': (port,)})
            )

        for args in add_nets:
            updaters.append(
                progress.ServerUpdateProgress(
                    self.resource_id, 'interface_attach',
                    handler_extra={'kwargs': args},
                    checker_extra={'args': (args['port_id'],)})
            )

        return updaters

    def needs_replace_with_prop_diff(self, changed_properties_set,
                                     after_props, before_props):
        """Needs replace based on prop_diff."""
        if self.FLAVOR in changed_properties_set:
            flavor_update_policy = (
                after_props.get(self.FLAVOR_UPDATE_POLICY) or
                before_props.get(self.FLAVOR_UPDATE_POLICY))
            if flavor_update_policy == 'REPLACE':
                return True

        if self.IMAGE in changed_properties_set:
            image_update_policy = (
                after_props.get(self.IMAGE_UPDATE_POLICY) or
                before_props.get(self.IMAGE_UPDATE_POLICY))
            if image_update_policy == 'REPLACE':
                return True

        if self.USER_DATA in changed_properties_set:
            ud_update_policy = (
                after_props.get(self.USER_DATA_UPDATE_POLICY) or
                before_props.get(self.USER_DATA_UPDATE_POLICY))
            return ud_update_policy == 'REPLACE'

    def needs_replace_failed(self):
        if not self.resource_id:
            return True

        with self.client_plugin().ignore_not_found:
            server = self.client().servers.get(self.resource_id)
            return server.status in ('ERROR', 'DELETED', 'SOFT_DELETED')

        return True

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        updaters = super(Server, self).handle_update(
            json_snippet,
            tmpl_diff,
            prop_diff)
        server = None

        after_props = json_snippet.properties(self.properties_schema,
                                              self.context)
        if self.METADATA in prop_diff:
            server = self.client_plugin().get_server(self.resource_id)
            self.client_plugin().meta_update(server,
                                             after_props[self.METADATA])

        if self.TAGS in prop_diff:
            self._update_server_tags(after_props[self.TAGS] or [])

        if self.NAME in prop_diff:
            if not server:
                server = self.client_plugin().get_server(self.resource_id)
            self.client_plugin().rename(server, after_props[self.NAME])

        if self.NETWORKS in prop_diff:
            updaters.extend(self._update_networks(server, after_props))

        if self.FLAVOR in prop_diff:
            updaters.extend(self._update_flavor(after_props))

        if self.USER_DATA in prop_diff:
            # We only care about rebuild here. The standard replace is
            # dealt elsewere
            if after_props[self.USER_DATA_UPDATE_POLICY] == 'REBUILD':
                updaters.append(self._update_user_data_rebuild(after_props))

        if self.IMAGE in prop_diff:
            updaters.append(self._update_image(after_props))
        elif self.ADMIN_PASS in prop_diff:
            if not server:
                server = self.client_plugin().get_server(self.resource_id)
            server.change_password(after_props[self.ADMIN_PASS])

        # NOTE(pas-ha) optimization is possible (starting first task
        # right away), but we'd rather not, as this method already might
        # have called several APIs
        return updaters

    def check_update_complete(self, updaters):
        """Push all updaters to completion in list order."""
        for prg in updaters:
            if not prg.called:
                handler = getattr(self.client_plugin(), prg.handler)
                prg.called = handler(*prg.handler_args,
                                     **prg.handler_kwargs)
                return False
            if not prg.complete:
                check_complete = getattr(self.client_plugin(), prg.checker)
                prg.complete = check_complete(*prg.checker_args,
                                              **prg.checker_kwargs)
                break
        status = all(prg.complete for prg in updaters)
        if status:
            self.store_external_ports()
        return status

    def _validate_block_device_mapping(self):

        # either volume_id or snapshot_id needs to be specified, but not both
        # for block device mapping.
        bdm = self.properties[self.BLOCK_DEVICE_MAPPING] or []
        bdm_v2 = self.properties[self.BLOCK_DEVICE_MAPPING_V2] or []
        image = self.properties[self.IMAGE]
        if bdm and bdm_v2:
            raise exception.ResourcePropertyConflict(
                self.BLOCK_DEVICE_MAPPING, self.BLOCK_DEVICE_MAPPING_V2)
        bootable = image is not None
        for mapping in bdm:
            device_name = mapping[self.BLOCK_DEVICE_MAPPING_DEVICE_NAME]
            if device_name == 'vda':
                bootable = True

            volume_id = mapping.get(self.BLOCK_DEVICE_MAPPING_VOLUME_ID)
            snapshot_id = mapping.get(self.BLOCK_DEVICE_MAPPING_SNAPSHOT_ID)
            if volume_id is not None and snapshot_id is not None:
                raise exception.ResourcePropertyConflict(
                    self.BLOCK_DEVICE_MAPPING_VOLUME_ID,
                    self.BLOCK_DEVICE_MAPPING_SNAPSHOT_ID)
            if volume_id is None and snapshot_id is None:
                msg = _('Either volume_id or snapshot_id must be specified for'
                        ' device mapping %s') % device_name
                raise exception.StackValidationFailed(message=msg)

        bootable_devs = [image]
        for mapping in bdm_v2:
            volume_id = mapping.get(self.BLOCK_DEVICE_MAPPING_VOLUME_ID)
            snapshot_id = mapping.get(self.BLOCK_DEVICE_MAPPING_SNAPSHOT_ID)
            image_id = mapping.get(self.BLOCK_DEVICE_MAPPING_IMAGE)
            boot_index = mapping.get(self.BLOCK_DEVICE_MAPPING_BOOT_INDEX)
            swap_size = mapping.get(self.BLOCK_DEVICE_MAPPING_SWAP_SIZE)
            ephemeral = (mapping.get(
                self.BLOCK_DEVICE_MAPPING_EPHEMERAL_SIZE) or mapping.get(
                self.BLOCK_DEVICE_MAPPING_EPHEMERAL_FORMAT))

            property_tuple = (volume_id, snapshot_id, image_id, swap_size,
                              ephemeral)

            if property_tuple.count(None) < 4:
                raise exception.ResourcePropertyConflict(
                    self.BLOCK_DEVICE_MAPPING_VOLUME_ID,
                    self.BLOCK_DEVICE_MAPPING_SNAPSHOT_ID,
                    self.BLOCK_DEVICE_MAPPING_IMAGE,
                    self.BLOCK_DEVICE_MAPPING_SWAP_SIZE,
                    self.BLOCK_DEVICE_MAPPING_EPHEMERAL_SIZE,
                    self.BLOCK_DEVICE_MAPPING_EPHEMERAL_FORMAT
                )

            if property_tuple.count(None) == 5:
                msg = _('Either volume_id, snapshot_id, image_id, swap_size, '
                        'ephemeral_size or ephemeral_format must be '
                        'specified.')
                raise exception.StackValidationFailed(message=msg)

            if any((volume_id is not None, snapshot_id is not None,
                    image_id is not None)):
                # boot_index is not specified, set boot_index=0 when
                # build_block_device_mapping for volume, snapshot, image
                if boot_index is None or boot_index == 0:
                    bootable = True
                    bootable_devs.append(volume_id)
                    bootable_devs.append(snapshot_id)
                    bootable_devs.append(image_id)
        if not bootable:
            msg = _('Neither image nor bootable volume is specified for '
                    'instance %s') % self.name
            raise exception.StackValidationFailed(message=msg)
        if bdm_v2 and len(list(
                dev for dev in bootable_devs if dev is not None)) != 1:
            msg = _('Multiple bootable sources for instance %s.') % self.name
            raise exception.StackValidationFailed(message=msg)

    def _validate_image_flavor(self, image, flavor):
        try:
            image_obj = self.client_plugin('glance').get_image(image)
            flavor_obj = self.client_plugin().get_flavor(flavor)
        except Exception as ex:
            # Flavor or image may not have been created in the backend
            # yet when they are part of the same stack/template.
            if (self.client_plugin().is_not_found(ex) or
                    self.client_plugin('glance').is_not_found(ex)):
                return
            raise
        else:
            if image_obj.status.lower() != self.IMAGE_ACTIVE:
                msg = _('Image status is required to be %(cstatus)s not '
                        '%(wstatus)s.') % {
                    'cstatus': self.IMAGE_ACTIVE,
                    'wstatus': image_obj.status}
                raise exception.StackValidationFailed(message=msg)

            # validate image/flavor combination
            if flavor_obj.ram < image_obj.min_ram:
                msg = _('Image %(image)s requires %(imram)s minimum ram. '
                        'Flavor %(flavor)s has only %(flram)s.') % {
                    'image': image, 'imram': image_obj.min_ram,
                    'flavor': flavor, 'flram': flavor_obj.ram}
                raise exception.StackValidationFailed(message=msg)

            # validate image/flavor disk compatibility
            if flavor_obj.disk < image_obj.min_disk:
                msg = _('Image %(image)s requires %(imsz)s GB minimum '
                        'disk space. Flavor %(flavor)s has only '
                        '%(flsz)s GB.') % {
                    'image': image, 'imsz': image_obj.min_disk,
                    'flavor': flavor, 'flsz': flavor_obj.disk}
                raise exception.StackValidationFailed(message=msg)

    def validate(self):
        """Validate any of the provided params."""
        super(Server, self).validate()

        if self.user_data_software_config():
            if 'deployments' in self.t.metadata():
                msg = _('deployments key not allowed in resource metadata '
                        'with user_data_format of SOFTWARE_CONFIG')
                raise exception.StackValidationFailed(message=msg)

        self._validate_block_device_mapping()

        # make sure the image exists if specified.
        image = self.properties[self.IMAGE]
        flavor = self.properties[self.FLAVOR]
        if image:
            self._validate_image_flavor(image, flavor)

        networks = self.properties[self.NETWORKS] or []
        for network in networks:
            self._validate_network(network)

        has_str_net = self._str_network(networks) is not None
        if has_str_net:
            if len(networks) != 1:
                msg = _('Property "%s" can not be specified if '
                        'multiple network interfaces set for '
                        'server.') % self.ALLOCATE_NETWORK
                raise exception.StackValidationFailed(message=msg)
            # Check if str_network is allowed to use
            if not self.client_plugin().is_version_supported(
                    MICROVERSION_STR_NETWORK):
                msg = (_('Cannot use "%s" property - compute service '
                         'does not support the required api '
                         'microversion.') % self.ALLOCATE_NETWORK)
                raise exception.StackValidationFailed(message=msg)

        # record if any networks include explicit ports
        has_port = any(n[self.NETWORK_PORT] is not None for n in networks)
        # if 'security_groups' present for the server and explicit 'port'
        # in one or more entries in 'networks', raise validation error
        if has_port and self.properties[self.SECURITY_GROUPS]:
            raise exception.ResourcePropertyConflict(
                self.SECURITY_GROUPS,
                "/".join([self.NETWORKS, self.NETWORK_PORT]))

        # Check if nic tag is allowed to use
        if self._is_nic_tagged(networks=networks):
            if not self.client_plugin().is_version_supported(
                    MICROVERSION_NIC_TAGS):
                msg = (_('Cannot use "%s" property in networks - '
                         'nova does not support required '
                         'api microversion.'), self.NIC_TAG)
                raise exception.StackValidationFailed(message=msg)

        # Check if tags is allowed to use
        if self.properties[self.TAGS]:
            if not self.client_plugin().is_version_supported(
                    MICROVERSION_TAGS):
                msg = (_('Cannot use "%s" property - nova does not support '
                         'required api microversion.') % self.TAGS)
                raise exception.StackValidationFailed(message=msg)

        # retrieve provider's absolute limits if it will be needed
        metadata = self.properties[self.METADATA]
        personality = self.properties[self.PERSONALITY]

        if personality:
            if self.client_plugin().is_version_supported(
                    MICROVERSION_PERSONALITY_REMOVED):
                msg = (_('Cannot use the personality parameter as nova no '
                         'longer supports it. Use user_data instead.'))
                raise exception.StackValidationFailed(message=msg)

        if metadata or personality:
            limits = self.client_plugin().absolute_limits()

        # verify that the number of metadata entries is not greater
        # than the maximum number allowed in the provider's absolute
        # limits
        if metadata:
            msg = _('Instance metadata must not contain greater than %s '
                    'entries.  This is the maximum number allowed by your '
                    'service provider') % limits['maxServerMeta']
            self._check_maximum(len(metadata),
                                limits['maxServerMeta'], msg)

        # verify the number of personality files and the size of each
        # personality file against the provider's absolute limits
        if personality:
            msg = _("The personality property may not contain "
                    "greater than %s entries.") % limits['maxPersonality']
            self._check_maximum(len(personality),
                                limits['maxPersonality'], msg)

            for path, contents in personality.items():
                msg = (_("The contents of personality file \"%(path)s\" "
                         "is larger than the maximum allowed personality "
                         "file size (%(max_size)s bytes).") %
                       {'path': path,
                        'max_size': limits['maxPersonalitySize']})
                self._check_maximum(len(bytes(contents.encode('utf-8'))
                                        ) if contents is not None else 0,
                                    limits['maxPersonalitySize'], msg)

    def _delete(self):
        if self.user_data_software_config():
            self._delete_queue()
            self._delete_user()
            self._delete_temp_url()

        # remove internal and external ports
        self._delete_internal_ports()
        self.data_delete('external_ports')

        if self.resource_id is None:
            return

        try:
            self.client().servers.delete(self.resource_id)
        except Exception as e:
            self.client_plugin().ignore_not_found(e)
            return
        return progress.ServerDeleteProgress(self.resource_id)

    def handle_snapshot_delete(self, state):

        if state[1] != self.FAILED and self.resource_id:
            image_id = self.client().servers.create_image(
                self.resource_id, self.physical_resource_name())
            return progress.ServerDeleteProgress(
                self.resource_id, image_id, False)
        return self._delete()

    def handle_delete(self):

        return self._delete()

    def check_delete_complete(self, prg):
        if not prg:
            return True
        if not prg.image_complete:
            image = self.client_plugin('glance').get_image(prg.image_id)
            if image.status.lower() in (self.IMAGE_ERROR,
                                        self.IMAGE_DELETED):
                raise exception.Error(image.status)
            elif image.status.lower() == self.IMAGE_ACTIVE:
                prg.image_complete = True
                if not self._delete():
                    return True
            return False

        return self.client_plugin().check_delete_server_complete(
            prg.server_id)

    def handle_suspend(self):
        """Suspend a server.

        Note we do not wait for the SUSPENDED state, this is polled for by
        check_suspend_complete in a similar way to the create logic so we can
        take advantage of coroutines.
        """
        if self.resource_id is None:
            raise exception.Error(_('Cannot suspend %s, resource_id not set') %
                                  self.name)

        try:
            server = self.client().servers.get(self.resource_id)
        except Exception as e:
            if self.client_plugin().is_not_found(e):
                raise exception.NotFound(_('Failed to find server %s') %
                                         self.resource_id)
            else:
                raise
        else:
            # if the server has been suspended successful,
            # no need to suspend again
            if self.client_plugin().get_status(server) != 'SUSPENDED':
                LOG.debug('suspending server %s', self.resource_id)
                server.suspend()
            return server.id

    def check_suspend_complete(self, server_id):
        cp = self.client_plugin()
        server = cp.fetch_server(server_id)
        if not server:
            return False
        status = cp.get_status(server)
        LOG.debug('%(name)s check_suspend_complete status = %(status)s',
                  {'name': self.name, 'status': status})
        if status in (cp.deferred_server_statuses | {'ACTIVE'}):
            return status == 'SUSPENDED'
        else:
            exc = exception.ResourceUnknownStatus(
                result=_('Suspend of server %s failed') % server.name,
                resource_status=status)
            raise exc

    def handle_resume(self):
        """Resume a server.

        Note we do not wait for the ACTIVE state, this is polled for by
        check_resume_complete in a similar way to the create logic so we can
        take advantage of coroutines.
        """
        if self.resource_id is None:
            raise exception.Error(_('Cannot resume %s, resource_id not set') %
                                  self.name)

        try:
            server = self.client().servers.get(self.resource_id)
        except Exception as e:
            if self.client_plugin().is_not_found(e):
                raise exception.NotFound(_('Failed to find server %s') %
                                         self.resource_id)
            else:
                raise
        else:
            # if the server has been resumed successful,
            # no need to resume again
            if self.client_plugin().get_status(server) != 'ACTIVE':
                LOG.debug('resuming server %s', self.resource_id)
                server.resume()
            return server.id

    def check_resume_complete(self, server_id):
        return self.client_plugin()._check_active(server_id)

    def handle_snapshot(self):
        image_id = self.client().servers.create_image(
            self.resource_id, self.physical_resource_name())
        self.data_set('snapshot_image_id', image_id)
        return image_id

    def check_snapshot_complete(self, image_id):
        image = self.client_plugin('glance').get_image(image_id)
        if image.status.lower() == self.IMAGE_ACTIVE:
            server = self.client_plugin().get_server(self.resource_id)
            task_state = getattr(server, 'OS-EXT-STS:task_state', '')
            if task_state not in {'image_uploading', 'image_snapshot_pending',
                                  'image_snapshot', 'image_pending_upload'}:
                return True
        elif image.status.lower() in (self.IMAGE_ERROR, self.IMAGE_DELETED):
            raise exception.Error(image.status)

        return False

    def handle_delete_snapshot(self, snapshot):
        image_id = snapshot['resource_data'].get('snapshot_image_id')
        with self.client_plugin('glance').ignore_not_found:
            self.client('glance').images.delete(image_id)

    def handle_restore(self, defn, restore_data):
        image_id = restore_data['resource_data']['snapshot_image_id']
        props = dict((k, v) for k, v in self.properties.data.items()
                     if v is not None)
        for key in [self.BLOCK_DEVICE_MAPPING, self.BLOCK_DEVICE_MAPPING_V2,
                    self.NETWORKS]:
            if props.get(key) is not None:
                props[key] = list(dict((k, v) for k, v in prop.items()
                                       if v is not None)
                                  for prop in props[key])
        props[self.IMAGE] = image_id
        return defn.freeze(properties=props)

    def prepare_for_replace(self):
        # if the server has not been created yet, do nothing
        if self.resource_id is None:
            return

        self.prepare_ports_for_replace()

    def restore_prev_rsrc(self, convergence=False):
        self.restore_ports_after_rollback(convergence=convergence)


def resource_mapping():
    return {
        'OS::Nova::Server': Server,
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_messaging.rpc import dispatcher

from heat.common import exception
from heat.common import template_format
from heat.engine import service
from heat.engine import stack
from heat.engine import template as templatem
from heat.objects import stack as stack_object
from heat.tests import common
from heat.tests.engine import tools
from heat.tests import utils


class StackServiceActionsTest(common.HeatTestCase):

    def setUp(self):
        super(StackServiceActionsTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.man = service.EngineService('a-host', 'a-topic')
        self.man.thread_group_mgr = service.ThreadGroupManager()

    @mock.patch.object(stack.Stack, 'load')
    @mock.patch.object(service.ThreadGroupManager, 'start')
    def test_stack_suspend(self, mock_start, mock_load):
        stack_name = 'service_suspend_test_stack'
        t = template_format.parse(tools.wp_template)
        stk = utils.parse_stack(t, stack_name=stack_name)
        s = stack_object.Stack.get_by_id(self.ctx, stk.id)

        mock_load.return_value = stk
        thread = mock.MagicMock()
        mock_link = self.patchobject(thread, 'link')
        mock_start.return_value = thread
        self.patchobject(service, 'NotifyEvent')

        result = self.man.stack_suspend(self.ctx, stk.identifier())
        self.assertIsNone(result)
        mock_load.assert_called_once_with(self.ctx, stack=s)
        mock_link.assert_called_once_with(mock.ANY)
        mock_start.assert_called_once_with(stk.id, stk.suspend,
                                           notify=mock.ANY)

        stk.delete()

    @mock.patch.object(stack.Stack, 'load')
    @mock.patch.object(service.ThreadGroupManager, 'start')
    def test_stack_resume(self, mock_start, mock_load):
        stack_name = 'service_resume_test_stack'
        t = template_format.parse(tools.wp_template)
        stk = utils.parse_stack(t, stack_name=stack_name)

        mock_load.return_value = stk
        thread = mock.MagicMock()
        mock_link = self.patchobject(thread, 'link')
        mock_start.return_value = thread
        self.patchobject(service, 'NotifyEvent')

        result = self.man.stack_resume(self.ctx, stk.identifier())
        self.assertIsNone(result)

        mock_load.assert_called_once_with(self.ctx, stack=mock.ANY)
        mock_link.assert_called_once_with(mock.ANY)
        mock_start.assert_called_once_with(stk.id, stk.resume, notify=mock.ANY)

        stk.delete()

    def test_stack_suspend_nonexist(self):
        stack_name = 'service_suspend_nonexist_test_stack'
        t = template_format.parse(tools.wp_template)
        tmpl = templatem.Template(t)
        stk = stack.Stack(self.ctx, stack_name, tmpl)

        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.man.stack_suspend, self.ctx,
                               stk.identifier())
        self.assertEqual(exception.EntityNotFound, ex.exc_info[0])

    def test_stack_resume_nonexist(self):
        stack_name = 'service_resume_nonexist_test_stack'
        t = template_format.parse(tools.wp_template)
        tmpl = templatem.Template(t)
        stk = stack.Stack(self.ctx, stack_name, tmpl)

        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.man.stack_resume, self.ctx,
                               stk.identifier())
        self.assertEqual(exception.EntityNotFound, ex.exc_info[0])

    def _mock_thread_start(self, stack_id, func, *args, **kwargs):
        func(*args, **kwargs)
        return mock.Mock()

    @mock.patch.object(service.ThreadGroupManager, 'start')
    @mock.patch.object(stack.Stack, 'load')
    def test_stack_check(self, mock_load, mock_start):
        stack_name = 'service_check_test_stack'
        t = template_format.parse(tools.wp_template)
        stk = utils.parse_stack(t, stack_name=stack_name)

        stk.check = mock.Mock()
        self.patchobject(service, 'NotifyEvent')
        mock_load.return_value = stk
        mock_start.side_effect = self._mock_thread_start

        self.man.stack_check(self.ctx, stk.identifier())
        self.assertTrue(stk.check.called)

        stk.delete()


class StackServiceUpdateActionsNotSupportedTest(common.HeatTestCase):

    scenarios = [
        ('suspend_in_progress', dict(action='SUSPEND', status='IN_PROGRESS')),
        ('suspend_complete', dict(action='SUSPEND', status='COMPLETE')),
        ('suspend_failed', dict(action='SUSPEND', status='FAILED')),
        ('delete_in_progress', dict(action='DELETE', status='IN_PROGRESS')),
        ('delete_complete', dict(action='DELETE', status='COMPLETE')),
        ('delete_failed', dict(action='DELETE', status='FAILED')),
    ]

    def setUp(self):
        super(StackServiceUpdateActionsNotSupportedTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.man = service.EngineService('a-host', 'a-topic')

    @mock.patch.object(stack.Stack, 'load')
    def test_stack_update_actions_not_supported(self, mock_load):
        stack_name = '%s-%s' % (self.action, self.status)
        t = template_format.parse(tools.wp_template)
        old_stack = utils.parse_stack(t, stack_name=stack_name)

        old_stack.action = self.action
        old_stack.status = self.status

        sid = old_stack.store()
        s = stack_object.Stack.get_by_id(self.ctx, sid)

        mock_load.return_value = old_stack

        params = {'foo': 'bar'}
        template = '{ "Resources": {} }'
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.man.update_stack,
                               self.ctx, old_stack.identifier(), template,
                               params, None, {})
        self.assertEqual(exception.NotSupported, ex.exc_info[0])
        mock_load.assert_called_once_with(self.ctx, stack=s,
                                          check_refresh_cred=True)

        old_stack.delete()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from heat.scaling import template
from heat.tests import common


class ResourceTemplatesTest(common.HeatTestCase):

    def setUp(self):
        super(ResourceTemplatesTest, self).setUp()
        ids = ('stubbed-id-%s' % (i,) for i in itertools.count())
        self.next_id = lambda: next(ids)

    def test_create_template(self):
        """Test case for creating template.

        When creating a template from scratch, an empty list is accepted as
        the "old" resources and new resources are created up to num_resource.
        """
        templates = template.member_definitions([], {'type': 'Foo'}, 2, 0,
                                                self.next_id)
        expected = [
            ('stubbed-id-0', {'type': 'Foo'}),
            ('stubbed-id-1', {'type': 'Foo'})]
        self.assertEqual(expected, list(templates))

    def test_replace_template(self):
        """Test case for replacing template.

        If num_replace is the number of old resources, then all of the
        resources will be replaced.
        """
        old_resources = [
            ('old-id-0', {'type': 'Foo'}),
            ('old-id-1', {'type': 'Foo'})]
        templates = template.member_definitions(old_resources, {'type': 'Bar'},
                                                1, 2, self.next_id)
        expected = [('old-id-1', {'type': 'Bar'})]
        self.assertEqual(expected, list(templates))

    def test_replace_some_units(self):
        """Test case for making only the number of replacements specified.

        If the resource definition changes, only the number of replacements
        specified will be made; beyond that, the original templates are used.
        """
        old_resources = [
            ('old-id-0', {'type': 'Foo'}),
            ('old-id-1', {'type': 'Foo'})]
        new_spec = {'type': 'Bar'}
        templates = template.member_definitions(old_resources, new_spec, 2, 1,
                                                self.next_id)
        expected = [
            ('old-id-0', {'type': 'Bar'}),
            ('old-id-1', {'type': 'Foo'})]
        self.assertEqual(expected, list(templates))

    def test_growth_counts_as_replacement(self):
        """Test case for growing template.

        If we grow the template and replace some elements at the same time, the
        number of replacements to perform is reduced by the number of new
        resources to be created.
        """
        spec = {'type': 'Foo'}
        old_resources = [
            ('old-id-0', spec),
            ('old-id-1', spec)]
        new_spec = {'type': 'Bar'}
        templates = template.member_definitions(old_resources, new_spec, 4, 2,
                                                self.next_id)
        expected = [
            ('old-id-0', spec),
            ('old-id-1', spec),
            ('stubbed-id-0', new_spec),
            ('stubbed-id-1', new_spec)]
        self.assertEqual(expected, list(templates))

    def test_replace_units_some_already_up_to_date(self):
        """Test case for up-to-date resources in template.

        If some of the old resources already have the new resource definition,
        then they won't be considered for replacement, and the next resource
        that is out-of-date will be replaced.
        """
        old_resources = [
            ('old-id-0', {'type': 'Bar'}),
            ('old-id-1', {'type': 'Foo'})]
        new_spec = {'type': 'Bar'}
        templates = template.member_definitions(old_resources, new_spec, 2, 1,
                                                self.next_id)
        second_batch_expected = [
            ('old-id-0', {'type': 'Bar'}),
            ('old-id-1', {'type': 'Bar'})]
        self.assertEqual(second_batch_expected, list(templates))



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_messaging import exceptions
from webob import exc

from heat.api.openstack.v1 import util
from heat.common.i18n import _
from heat.common import serializers
from heat.common import wsgi
from heat.rpc import client as rpc_client


class ServiceController(object):
    """WSGI controller for reporting the heat engine status in Heat v1 API."""
    # Define request scope (must match what is in policy.yaml or policies in
    # code)
    REQUEST_SCOPE = 'service'

    def __init__(self, options):
        self.options = options
        self.rpc_client = rpc_client.EngineClient()

    @util.registered_policy_enforce
    def index(self, req):
        try:
            services = self.rpc_client.list_services(req.context)
            return {'services': services}
        except exceptions.MessagingTimeout:
            msg = _('All heat engines are down.')
            raise exc.HTTPServiceUnavailable(msg)


def create_resource(options):
    deserializer = wsgi.JSONRequestDeserializer()
    serializer = serializers.JSONResponseSerializer()
    return wsgi.Resource(ServiceController(options), deserializer, serializer)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from urllib import parse as urlparse

from heat.common import template_format
from heat.engine.clients import client_plugin
from heat.engine import resource
from heat.engine.resources.openstack.zaqar import queue
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

try:
    from zaqarclient.transport.errors import ResourceNotFound  # noqa
except ImportError:
    ResourceNotFound = Exception

wp_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "openstack Zaqar queue service as a resource",
  "Resources" : {
    "MyQueue2" : {
      "Type" : "OS::Zaqar::Queue",
      "Properties" : {
        "name": "myqueue",
        "metadata": { "key1": { "key2": "value", "key3": [1, 2] } }
      }
    }
  },
  "Outputs" : {
    "queue_id": {
      "Value": { "Ref" : "MyQueue2" },
      "Description": "queue name"
    },
    "queue_href": {
      "Value": { "Fn::GetAtt" : [ "MyQueue2", "href" ]},
      "Description": "queue href"
    }
  }
}
'''


class FakeQueue(object):
    def __init__(self, queue_name, auto_create=True):
        self._id = queue_name
        self._auto_create = auto_create
        self._exists = False
        self._metadata = None
        self._delete_called = False
        self.metadata = mock.Mock()
        self.delete = mock.Mock()


class FakeClient(object):
    def __init__(self):
        self.api_url = 'http://127.0.0.1:8888/'
        self.api_version = 1.1
        self.queue = mock.Mock()


class ZaqarMessageQueueTest(common.HeatTestCase):
    def setUp(self):
        super(ZaqarMessageQueueTest, self).setUp()
        self.fc = FakeClient()
        self.patchobject(resource.Resource, 'client', return_value=self.fc)
        self.ctx = utils.dummy_context()

    def parse_stack(self, t):
        stack_name = 'test_stack'
        tmpl = template.Template(t)
        self.stack = stack.Stack(self.ctx, stack_name, tmpl)
        self.stack.validate()
        self.stack.store()

    def test_create(self):
        t = template_format.parse(wp_template)
        self.parse_stack(t)

        queue = self.stack['MyQueue2']
        queue_metadata = queue.properties.get('metadata')
        fake_q = FakeQueue(queue.physical_resource_name(), auto_create=False)
        fake_q.metadata.return_value = queue_metadata
        self.fc.queue.return_value = fake_q

        scheduler.TaskRunner(queue.create)()
        self.assertEqual('http://127.0.0.1:8888/v1.1/queues/myqueue',
                         queue.FnGetAtt('href'))
        self.fc.queue.assert_called_once_with(queue.physical_resource_name(),
                                              auto_create=False)
        fake_q.metadata.assert_called_once_with(new_meta=queue_metadata)

    def test_create_default_name(self):
        t = template_format.parse(wp_template)
        del t['Resources']['MyQueue2']['Properties']['name']
        self.parse_stack(t)

        queue = self.stack['MyQueue2']

        name_match = utils.PhysName(self.stack.name, 'MyQueue2')
        self.fc.queue.side_effect = FakeQueue
        scheduler.TaskRunner(queue.create)()
        queue_name = queue.physical_resource_name()
        self.assertEqual(name_match, queue_name)

        self.fc.api_version = 2
        self.assertEqual('http://127.0.0.1:8888/v2/queues/' + queue_name,
                         queue.FnGetAtt('href'))
        self.fc.queue.assert_called_once_with(name_match, auto_create=False)

    def test_delete(self):
        t = template_format.parse(wp_template)
        self.parse_stack(t)

        queue = self.stack['MyQueue2']
        queue.resource_id_set(queue.properties.get('name'))

        fake_q = FakeQueue("myqueue", auto_create=False)
        self.fc.queue.return_value = fake_q

        scheduler.TaskRunner(queue.create)()
        self.fc.queue.assert_called_once_with("myqueue", auto_create=False)
        scheduler.TaskRunner(queue.delete)()
        fake_q.delete.assert_called()

    @mock.patch.object(queue.ZaqarQueue, "client")
    def test_delete_not_found(self, mockclient):
        class ZaqarClientPlugin(client_plugin.ClientPlugin):
            def _create(self):
                return mockclient()

        mock_def = mock.Mock(spec=rsrc_defn.ResourceDefinition)
        mock_def.resource_type = 'OS::Zaqar::Queue'
        props = mock.Mock()
        props.props = {}
        mock_def.properties.return_value = props
        stack = utils.parse_stack(template_format.parse(wp_template))
        self.patchobject(stack, 'db_resource_get', return_value=None)
        mockplugin = ZaqarClientPlugin(self.ctx)
        clients = self.patchobject(stack, 'clients')
        clients.client_plugin.return_value = mockplugin

        mockplugin.is_not_found = mock.Mock()
        mockplugin.is_not_found.return_value = True

        zaqar_q = mock.Mock()
        zaqar_q.delete.side_effect = ResourceNotFound()
        mockclient.return_value.queue.return_value = zaqar_q
        zplugin = queue.ZaqarQueue("test_delete_not_found", mock_def,
                                   stack)
        zplugin.resource_id = "test_delete_not_found"
        zplugin.handle_delete()
        clients.client_plugin.assert_called_once_with('zaqar')
        mockplugin.is_not_found.assert_called_once_with(
            zaqar_q.delete.side_effect)
        mockclient.return_value.queue.assert_called_once_with(
            "test_delete_not_found", auto_create=False)

    def test_update_in_place(self):
        t = template_format.parse(wp_template)
        self.parse_stack(t)
        queue = self.stack['MyQueue2']
        queue.resource_id_set(queue.properties.get('name'))
        fake_q = FakeQueue('myqueue', auto_create=False)
        self.fc.queue.return_value = fake_q
        t = template_format.parse(wp_template)
        new_queue = t['Resources']['MyQueue2']
        new_queue['Properties']['metadata'] = {'key1': 'value'}
        resource_defns = template.Template(t).resource_definitions(self.stack)

        scheduler.TaskRunner(queue.create)()
        self.fc.queue.assert_called_once_with("myqueue", auto_create=False)
        fake_q.metadata.assert_called_with(new_meta={'key1': {'key2': 'value',
                                                              'key3': [1, 2]}})

        scheduler.TaskRunner(queue.update, resource_defns['MyQueue2'])()
        fake_q.metadata.assert_called_with(
            new_meta={'key1': 'value'})

    def test_update_replace(self):
        t = template_format.parse(wp_template)
        self.parse_stack(t)
        queue = self.stack['MyQueue2']
        queue.resource_id_set(queue.properties.get('name'))
        fake_q = FakeQueue('myqueue', auto_create=False)
        self.fc.queue.return_value = fake_q

        t = template_format.parse(wp_template)
        t['Resources']['MyQueue2']['Properties']['name'] = 'new_queue'
        resource_defns = template.Template(t).resource_definitions(self.stack)
        new_queue = resource_defns['MyQueue2']

        scheduler.TaskRunner(queue.create)()
        self.fc.queue.assert_called_once_with("myqueue", auto_create=False)
        err = self.assertRaises(resource.UpdateReplace,
                                scheduler.TaskRunner(queue.update,
                                                     new_queue))
        msg = 'The Resource MyQueue2 requires replacement.'
        self.assertEqual(msg, str(err))

    def test_show_resource(self):
        t = template_format.parse(wp_template)
        self.parse_stack(t)

        queue = self.stack['MyQueue2']
        fake_q = FakeQueue(queue.physical_resource_name(), auto_create=False)
        queue_metadata = queue.properties.get('metadata')
        fake_q.metadata.return_value = queue_metadata
        self.fc.queue.return_value = fake_q

        scheduler.TaskRunner(queue.create)()
        self.fc.queue.assert_called_once_with(queue.physical_resource_name(),
                                              auto_create=False)
        self.assertEqual(
            {'metadata': {"key1": {"key2": "value", "key3": [1, 2]}}},
            queue._show_resource())
        fake_q.metadata.assert_called_with()

    def test_parse_live_resource_data(self):
        t = template_format.parse(wp_template)
        self.parse_stack(t)

        queue = self.stack['MyQueue2']
        fake_q = FakeQueue(queue.physical_resource_name(), auto_create=False)
        self.fc.queue.return_value = fake_q
        queue_metadata = queue.properties.get('metadata')
        fake_q.metadata.return_value = queue_metadata
        scheduler.TaskRunner(queue.create)()
        fake_q.metadata.assert_called_with(new_meta=queue_metadata)
        self.fc.queue.assert_called_once_with(queue.physical_resource_name(),
                                              auto_create=False)
        self.assertEqual(
            {'metadata': {"key1": {"key2": "value", "key3": [1, 2]}},
             'name': queue.resource_id},
            queue.parse_live_resource_data(queue.properties,
                                           queue._show_resource()))
        fake_q.metadata.assert_called_with()


class ZaqarSignedQueueURLTest(common.HeatTestCase):
    tmpl = '''
heat_template_version: 2015-10-15
resources:
  signed_url:
    type: OS::Zaqar::SignedQueueURL
    properties:
      queue: foo
      ttl: 60
      paths:
        - messages
        - subscription
      methods:
        - POST
        - DELETE
'''

    @mock.patch('zaqarclient.queues.v2.queues.Queue.signed_url')
    def test_create(self, mock_signed_url):
        mock_signed_url.return_value = {
            'expires': '2020-01-01',
            'signature': 'secret',
            'project': 'project_id',
            'paths': ['/v2/foo/messages', '/v2/foo/sub'],
            'methods': ['DELETE', 'POST']}

        self.t = template_format.parse(self.tmpl)
        self.stack = utils.parse_stack(self.t)
        self.rsrc = self.stack['signed_url']
        self.assertIsNone(self.rsrc.validate())
        self.stack.create()
        self.assertEqual(self.rsrc.CREATE, self.rsrc.action)
        self.assertEqual(self.rsrc.COMPLETE, self.rsrc.status)
        self.assertEqual(self.stack.CREATE, self.stack.action)
        self.assertEqual(self.stack.COMPLETE, self.stack.status)

        mock_signed_url.assert_called_once_with(
            paths=['messages', 'subscription'],
            methods=['POST', 'DELETE'],
            ttl_seconds=60)

        self.assertEqual('secret', self.rsrc.FnGetAtt('signature'))
        self.assertEqual('2020-01-01', self.rsrc.FnGetAtt('expires'))
        self.assertEqual('project_id', self.rsrc.FnGetAtt('project'))
        self.assertEqual(['/v2/foo/messages', '/v2/foo/sub'],
                         self.rsrc.FnGetAtt('paths'))
        self.assertEqual(['DELETE', 'POST'], self.rsrc.FnGetAtt('methods'))
        expected_query = {
            'queue_name': ['foo'],
            'expires': ['2020-01-01'],
            'signature': ['secret'],
            'project_id': ['project_id'],
            'paths': ['/v2/foo/messages,/v2/foo/sub'],
            'methods': ['DELETE,POST']
        }
        query_str_attr = self.rsrc.get_attribute('query_str')
        self.assertEqual(expected_query,
                         urlparse.parse_qs(query_str_attr,
                                           strict_parsing=True))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

from oslo_config import cfg
from troveclient import exceptions as troveexc
from troveclient.v1 import users

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine.clients.os import trove
from heat.engine import resource
from heat.engine.resources.openstack.trove import instance as dbinstance
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.engine import stack as parser
from heat.engine import template as tmpl
from heat.tests import common
from heat.tests import utils


db_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "MySQL instance running on openstack DBaaS cloud",
  "Resources" : {
    "MySqlCloudDB": {
      "Type": "OS::Trove::Instance",
      "Properties" : {
        "name" : "test",
        "flavor" : "1GB",
        "size" : 30,
        "users" : [{"name": "testuser", "password": "pass", "databases":
        ["validdb"]}],
        "databases" : [{"name": "validdb"}],
        "datastore_type": "SomeDStype",
        "datastore_version": "MariaDB-5.5"
      }
    }
  }

}
'''

db_template_with_nics = '''
heat_template_version: 2013-05-23
description: MySQL instance running on openstack DBaaS cloud
resources:
  MySqlCloudDB:
    type: OS::Trove::Instance
    properties:
      name: test
      flavor: 1GB
      size: 30
      networks:
        - port: someportname
          fixed_ip: 1.2.3.4
'''

db_template_with_replication = '''
heat_template_version: 2013-05-23
description: MySQL instance running on openstack DBaaS cloud
resources:
  MySqlCloudDB:
    type: OS::Trove::Instance
    properties:
      name: test
      flavor: 1GB
      size: 30
      replica_of: 0e642916-dd64-43b3-933f-ff34fff69a7f
      replica_count: 2
'''


class FakeDBInstance(object):
    def __init__(self):
        self.id = 12345
        self.hostname = "testhost"
        self.links = [
            {"href": "https://adga23dd432a.rackspacecloud.com/132345245",
             "rel": "self"}]
        self.resource_id = 12345
        self.status = 'ACTIVE'

    def delete(self):
        pass

    def to_dict(self):
        pass


class FakeFlavor(object):
    def __init__(self, id, name):
        self.id = id
        self.name = name


class FakeVersion(object):
    def __init__(self, name="MariaDB-5.5"):
        self.name = name


class InstanceTest(common.HeatTestCase):

    def setUp(self):
        super(InstanceTest, self).setUp()
        self.fc = mock.MagicMock()
        self.nova = mock.Mock()
        self.client = mock.Mock()
        self.patchobject(trove.TroveClientPlugin, '_create',
                         return_value=self.client)
        self.stub_TroveFlavorConstraint_validate()
        self.flavor_resolve = self.patchobject(trove.TroveClientPlugin,
                                               'find_flavor_by_name_or_id',
                                               return_value='1')
        self.fake_instance = FakeDBInstance()
        self.client.instances.create.return_value = self.fake_instance
        self.client.instances.get.return_value = self.fake_instance

    def _setup_test_instance(self, name, t, rsrc_name='MySqlCloudDB'):
        stack_name = '%s_stack' % name
        template = tmpl.Template(t)
        self.stack = parser.Stack(utils.dummy_context(),
                                  stack_name,
                                  template,
                                  stack_id=str(uuid.uuid4()))
        rsrc = self.stack[rsrc_name]
        rsrc.resource_id = '12345'
        return rsrc

    def _stubout_validate(self, instance, mock_net_constraint=False,
                          with_port=True):
        if mock_net_constraint:
            self.stub_NetworkConstraint_validate()

        self.client.datastore_versions.list.return_value = [FakeVersion()]
        if with_port:
            self.stub_PortConstraint_validate()

    def test_instance_create(self):
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('dbinstance_create', t)
        scheduler.TaskRunner(instance.create)()
        self.assertEqual((instance.CREATE, instance.COMPLETE), instance.state)
        self.assertEqual('instances', instance.entity)

    def test_create_failed(self):
        t = template_format.parse(db_template)
        osdb_res = self._setup_test_instance('dbinstance_create', t)

        trove_mock = mock.Mock()
        self.patchobject(osdb_res, 'client', return_value=trove_mock)

        # test for bad statuses
        mock_input = mock.Mock()
        mock_input.status = 'ERROR'
        trove_mock.instances.get.return_value = mock_input
        error_string = ('Went to status ERROR due to "The last operation for '
                        'the database instance failed due to an error."')
        exc = self.assertRaises(exception.ResourceInError,
                                osdb_res.check_create_complete,
                                mock_input)
        self.assertIn(error_string, str(exc))

        mock_input = mock.Mock()
        mock_input.status = 'FAILED'
        trove_mock.instances.get.return_value = mock_input
        error_string = ('Went to status FAILED due to "The database instance '
                        'was created, but heat failed to set up the '
                        'datastore. If a database instance is in the FAILED '
                        'state, it should be deleted and a new one should '
                        'be created."')
        exc = self.assertRaises(exception.ResourceInError,
                                osdb_res.check_create_complete,
                                mock_input)
        self.assertIn(error_string, str(exc))

        # test if error string is not defined

        osdb_res.TROVE_STATUS_REASON = {}
        mock_input = mock.Mock()
        mock_input.status = 'ERROR'
        error_string = ('Went to status ERROR due to "Unknown"')
        trove_mock.instances.get.return_value = mock_input
        exc = self.assertRaises(exception.ResourceInError,
                                osdb_res.check_create_complete,
                                mock_input)
        self.assertIn(error_string, str(exc))

    def _create_failed_bad_status(self, status, error_message):
        t = template_format.parse(db_template)
        bad_instance = mock.Mock()
        bad_instance.status = status
        self.client.instances.get.return_value = bad_instance
        instance = self._setup_test_instance('test_bad_statuses', t)
        ex = self.assertRaises(exception.ResourceInError,
                               instance.check_create_complete,
                               self.fake_instance.id)
        self.assertIn(error_message, str(ex))

    def test_create_failed_status_error(self):
        self._create_failed_bad_status(
            'ERROR', 'Went to status ERROR due to "The last operation for '
                     'the database instance failed due to an error."')

    def test_create_failed_status_failed(self):
        self._create_failed_bad_status(
            'FAILED', 'Went to status FAILED due to "The database instance '
                      'was created, but heat failed to set up the datastore. '
                      'If a database instance is in the FAILED state, it '
                      'should be deleted and a new one should be created."')

    def test_instance_restore_point(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties']['restore_point'] = "1234"
        instance = self._setup_test_instance('dbinstance_create', t)

        self.client.flavors.get.side_effect = [troveexc.NotFound()]
        self.client.flavors.find.return_value = FakeFlavor(1, '1GB')

        scheduler.TaskRunner(instance.create)()
        self.assertEqual((instance.CREATE, instance.COMPLETE), instance.state)

        users = [{"name": "testuser", "password": "pass", "host": "%",
                  "databases": [{"name": "validdb"}]}]
        databases = [{"collate": "utf8_general_ci",
                      "character_set": "utf8",
                      "name": "validdb"}]
        self.client.instances.create.assert_called_once_with(
            'test', '1', volume={'size': 30}, databases=databases, users=users,
            restorePoint={"backupRef": "1234"}, availability_zone=None,
            datastore="SomeDStype", datastore_version="MariaDB-5.5", nics=[],
            replica_of=None, replica_count=None)

    def test_instance_create_overlimit(self):
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('dbinstance_create', t)

        # Simulate an OverLimit exception
        self.client.instances.get.side_effect = [
            troveexc.RequestEntityTooLarge(), self.fake_instance]

        scheduler.TaskRunner(instance.create)()
        self.assertEqual((instance.CREATE, instance.COMPLETE), instance.state)

    def test_instance_create_fails(self):
        cfg.CONF.set_override('action_retry_limit', 0)
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('dbinstance_create', t)
        self.fake_instance.status = 'ERROR'
        self.assertRaises(exception.ResourceFailure,
                          scheduler.TaskRunner(instance.create))
        # return previous status
        self.fake_instance.status = 'ACTIVE'

    def _get_db_instance(self):
        t = template_format.parse(db_template)
        res = self._setup_test_instance('trove_check', t)
        res.state_set(res.CREATE, res.COMPLETE)
        res.flavor = 'Foo Flavor'
        res.volume = 'Foo Volume'
        res.datastore_type = 'Foo Type'
        res.datastore_version = 'Foo Version'
        return res

    def test_instance_check(self):
        res = self._get_db_instance()
        scheduler.TaskRunner(res.check)()
        self.assertEqual((res.CHECK, res.COMPLETE), res.state)

    def test_instance_check_not_active(self):
        res = self._get_db_instance()
        self.fake_instance.status = 'FOOBAR'

        exc = self.assertRaises(exception.ResourceFailure,
                                scheduler.TaskRunner(res.check))
        self.assertIn('FOOBAR', str(exc))
        self.assertEqual((res.CHECK, res.FAILED), res.state)
        # return previous status
        self.fake_instance.status = 'ACTIVE'

    def test_instance_delete(self):
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('dbinstance_del', t)
        self.client.instances.get.side_effect = [self.fake_instance,
                                                 troveexc.NotFound(404)]

        scheduler.TaskRunner(instance.create)()
        scheduler.TaskRunner(instance.delete)()

    def test_instance_delete_overlimit(self):
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('dbinstance_del', t)

        # Simulate an OverLimit exception
        self.client.instances.get.side_effect = [
            troveexc.RequestEntityTooLarge(), self.fake_instance,
            troveexc.NotFound(404)]

        scheduler.TaskRunner(instance.create)()
        scheduler.TaskRunner(instance.delete)()

    def test_instance_delete_resource_none(self):
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('dbinstance_del', t)
        scheduler.TaskRunner(instance.create)()
        instance.resource_id = None
        scheduler.TaskRunner(instance.delete)()
        self.assertIsNone(instance.resource_id)

    def test_instance_resource_not_found(self):
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('dbinstance_del', t)
        self.client.instances.get.side_effect = [self.fake_instance,
                                                 troveexc.NotFound(404)]
        scheduler.TaskRunner(instance.create)()
        scheduler.TaskRunner(instance.delete)()

    def test_instance_attributes(self):
        fake_instance = FakeDBInstance()
        self.client.instances.create.return_value = fake_instance
        self.client.instances.get.return_value = fake_instance
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('attr_test', t)
        self.assertEqual("testhost", instance.FnGetAtt('hostname'))
        self.assertEqual("https://adga23dd432a.rackspacecloud.com/132345245",
                         instance.FnGetAtt('href'))

    def test_instance_validation_success(self):
        t = template_format.parse(db_template)
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance)
        self.assertIsNone(instance.validate())

    def test_instance_validation_invalid_db(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties']['databases'] = [
            {"name": "onedb"}]
        t['Resources']['MySqlCloudDB']['Properties']['users'] = [
            {"name": "testuser",
             "password": "pass",
             "databases": ["invaliddb"]}]
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance)
        ex = self.assertRaises(exception.StackValidationFailed,
                               instance.validate)
        self.assertEqual("Database ['invaliddb'] specified for user does not "
                         "exist in databases for resource MySqlCloudDB.",
                         str(ex))

    def test_instance_validation_db_name_hyphens(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties']['databases'] = [
            {"name": "-foo-bar-"}]
        t['Resources']['MySqlCloudDB']['Properties']['users'] = [
            {"name": "testuser",
             "password": "pass",
             "databases": ["-foo-bar-"]}]
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance)
        self.assertIsNone(instance.validate())

    def test_instance_validation_users_none(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties']['users'] = []
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance)
        self.assertIsNone(instance.validate())

    def test_instance_validation_databases_none(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties']['databases'] = []
        t['Resources']['MySqlCloudDB']['Properties']['users'] = [
            {"name": "testuser",
             "password": "pass",
             "databases": ["invaliddb"]}]
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance)
        ex = self.assertRaises(exception.StackValidationFailed,
                               instance.validate)
        self.assertEqual('Databases property is required if users property '
                         'is provided for resource MySqlCloudDB.',
                         str(ex))

    def test_instance_validation_user_no_db(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties']['databases'] = [
            {"name": "validdb"}]
        t['Resources']['MySqlCloudDB']['Properties']['users'] = [
            {"name": "testuser", "password": "pass", "databases": []}]

        instance = self._setup_test_instance('dbinstance_test', t)
        ex = self.assertRaises(exception.StackValidationFailed,
                               instance.validate)
        self.assertEqual('Property error: '
                         'Resources.MySqlCloudDB.Properties.'
                         'users[0].databases: length (0) is out of range '
                         '(min: 1, max: None)', str(ex))

    def test_instance_validation_no_datastore_yes_version(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties'].pop('datastore_type')
        instance = self._setup_test_instance('dbinstance_test', t)
        ex = self.assertRaises(exception.StackValidationFailed,
                               instance.validate)
        exp_msg = "Not allowed - datastore_version without datastore_type."
        self.assertEqual(exp_msg, str(ex))

    def test_instance_validation_no_ds_version(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties'][
            'datastore_type'] = 'mysql'
        t['Resources']['MySqlCloudDB']['Properties'].pop('datastore_version')
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance)
        self.assertIsNone(instance.validate())

    def test_instance_validation_wrong_dsversion(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties'][
            'datastore_type'] = 'mysql'
        t['Resources']['MySqlCloudDB']['Properties'][
            'datastore_version'] = 'SomeVersion'
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance)

        ex = self.assertRaises(exception.StackValidationFailed,
                               instance.validate)
        expected_msg = ("Datastore version SomeVersion for datastore type "
                        "mysql is not valid. "
                        "Allowed versions are MariaDB-5.5.")
        self.assertEqual(expected_msg, str(ex))

    def test_instance_validation_implicit_version(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties'][
            'datastore_type'] = 'mysql'
        t['Resources']['MySqlCloudDB']['Properties'].pop('datastore_version')
        instance = self._setup_test_instance('dbinstance_test', t)
        self.client.datastore_versions.list.return_value = [
            FakeVersion(), FakeVersion('MariaDB-5.0')]
        self.assertIsNone(instance.validate())

    def test_instance_validation_net_with_port_fail(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties']['networks'] = [
            {
                "port": "someportuuid",
                "network": "somenetuuid"
            }]
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance, mock_net_constraint=True)

        ex = self.assertRaises(
            exception.StackValidationFailed, instance.validate)
        self.assertEqual('Either network or port must be provided.',
                         str(ex))

    def test_instance_validation_no_net_no_port_fail(self):
        t = template_format.parse(db_template)
        t['Resources']['MySqlCloudDB']['Properties']['networks'] = [
            {
                "fixed_ip": "1.2.3.4"
            }]
        instance = self._setup_test_instance('dbinstance_test', t)
        self._stubout_validate(instance, with_port=False)

        ex = self.assertRaises(
            exception.StackValidationFailed, instance.validate)
        self.assertEqual('Either network or port must be provided.',
                         str(ex))

    def test_instance_create_with_port(self):
        t = template_format.parse(db_template_with_nics)
        instance = self._setup_test_instance('dbinstance_test', t)
        self.patchobject(neutron.NeutronClientPlugin,
                         'find_resourceid_by_name_or_id',
                         return_value='someportid')
        self.stub_PortConstraint_validate()

        scheduler.TaskRunner(instance.create)()
        self.assertEqual((instance.CREATE, instance.COMPLETE), instance.state)
        self.client.instances.create.assert_called_once_with(
            'test', '1', volume={'size': 30}, databases=[], users=[],
            restorePoint=None, availability_zone=None, datastore=None,
            datastore_version=None, nics=[{'port-id': 'someportid',
                                           'v4-fixed-ip': '1.2.3.4'}],
            replica_of=None, replica_count=None)

    def test_instance_create_with_net_id(self):
        net_id = '034aa4d5-0f36-4127-8481-5caa5bfc9403'
        t = template_format.parse(db_template_with_nics)
        t['resources']['MySqlCloudDB']['properties']['networks'] = [
            {'network': net_id}]
        instance = self._setup_test_instance('dbinstance_test', t)
        self.stub_NetworkConstraint_validate()
        self.patchobject(neutron.NeutronClientPlugin,
                         'find_resourceid_by_name_or_id',
                         return_value=net_id)
        scheduler.TaskRunner(instance.create)()
        self.assertEqual((instance.CREATE, instance.COMPLETE), instance.state)
        self.client.instances.create.assert_called_once_with(
            'test', '1', volume={'size': 30}, databases=[], users=[],
            restorePoint=None, availability_zone=None, datastore=None,
            datastore_version=None, nics=[{'net-id': net_id}], replica_of=None,
            replica_count=None)

    def test_instance_create_with_replication(self):
        t = template_format.parse(db_template_with_replication)
        instance = self._setup_test_instance('dbinstance_test', t)

        scheduler.TaskRunner(instance.create)()
        self.assertEqual((instance.CREATE, instance.COMPLETE), instance.state)
        self.client.instances.create.assert_called_once_with(
            'test', '1', volume={'size': 30}, databases=[], users=[],
            restorePoint=None, availability_zone=None, datastore=None,
            datastore_version=None, nics=[],
            replica_of="0e642916-dd64-43b3-933f-ff34fff69a7f", replica_count=2)

    def test_instance_get_live_state(self):
        self.fake_instance.to_dict = mock.Mock(return_value={
            'name': 'test_instance',
            'flavor': {'id': '1'},
            'volume': {'size': 30}
        })

        fake_db1 = mock.Mock()
        fake_db1.name = 'validdb'
        fake_db2 = mock.Mock()
        fake_db2.name = 'secondvaliddb'
        self.client.databases.list.return_value = [fake_db1, fake_db2]

        expected = {
            'flavor': '1',
            'name': 'test_instance',
            'size': 30,
            'databases': [{'name': 'validdb',
                           'character_set': 'utf8',
                           'collate': 'utf8_general_ci'},
                          {'name': 'secondvaliddb'}]
        }

        t = template_format.parse(db_template)
        instance = self._setup_test_instance('get_live_state_test', t)
        reality = instance.get_live_state(instance.properties)

        self.assertEqual(expected, reality)


@mock.patch.object(resource.Resource, "client_plugin")
@mock.patch.object(resource.Resource, "client")
class InstanceUpdateTests(common.HeatTestCase):

    def setUp(self):
        super(InstanceUpdateTests, self).setUp()
        self._stack = utils.parse_stack(template_format.parse(db_template))
        testprops = {
            "name": "testinstance",
            "flavor": "foo",
            "datastore_type": "database",
            "datastore_version": "1",
            "size": 10,
            "databases": [
                {"name": "bar"},
                {"name": "biff"}
            ],
            "users": [
                {
                    "name": "baz",
                    "password": "password",
                    "databases": ["bar"]
                },
                {
                    "name": "deleted",
                    "password": "password",
                    "databases": ["biff"]
                }
            ]
        }
        self._rdef = rsrc_defn.ResourceDefinition('test',
                                                  dbinstance.Instance,
                                                  properties=testprops)

    def test_handle_no_update(self, mock_client, mock_plugin):
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertEqual({}, trove.handle_update(None, None, {}))

    def test_handle_update_name(self, mock_client, mock_plugin):
        prop_diff = {
            "name": "changed"
        }
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertEqual(prop_diff, trove.handle_update(None, None, prop_diff))

    def test_handle_update_databases(self, mock_client, mock_plugin):
        prop_diff = {
            "databases": [
                {"name": "bar",
                 "character_set": "ascii"},
                {'name': "baz"}
            ]
        }
        mget = mock_client().databases.list
        mbar = mock.Mock(name='bar')
        mbar.name = 'bar'
        mbiff = mock.Mock(name='biff')
        mbiff.name = 'biff'
        mget.return_value = [mbar, mbiff]
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        expected = {
            'databases': [
                {'character_set': 'ascii', 'name': 'bar'},
                {'ACTION': 'CREATE', 'name': 'baz'},
                {'ACTION': 'DELETE', 'name': 'biff'}
            ]}
        self.assertEqual(expected, trove.handle_update(None, None, prop_diff))

    def test_handle_update_users(self, mock_client, mock_plugin):
        prop_diff = {
            "users": [
                {"name": "baz",
                 "password": "changed",
                 "databases": ["bar", "biff"]},
                {'name': "user2",
                 "password": "password",
                 "databases": ["biff", "bar"]}
            ]
        }
        uget = mock_client().users
        mbaz = mock.Mock(name='baz')
        mbaz.name = 'baz'
        mdel = mock.Mock(name='deleted')
        mdel.name = 'deleted'
        uget.list.return_value = [mbaz, mdel]
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        expected = {
            'users': [{
                'databases': ['bar', 'biff'],
                'name': 'baz',
                'password': 'changed'
            }, {
                'ACTION': 'CREATE',
                'databases': ['biff', 'bar'],
                'name': 'user2',
                'password': 'password'
            }, {
                'ACTION': 'DELETE',
                'name': 'deleted'
            }]}
        self.assertEqual(expected, trove.handle_update(None, None, prop_diff))

    def test_handle_update_flavor(self, mock_client, mock_plugin):
        # Translation mechanism already resolved flavor name to id.
        prop_diff = {
            "flavor": 1234
        }
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        expected = {
            "flavor": 1234
        }
        self.assertEqual(expected, trove.handle_update(None, None, prop_diff))

    def test_handle_update_size(self, mock_client, mock_plugin):
        prop_diff = {
            "size": 42
        }
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        expected = {
            "size": 42
        }
        self.assertEqual(expected, trove.handle_update(None, None, prop_diff))

    def test_check_complete_none(self, mock_client, mock_plugin):
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertTrue(trove.check_update_complete({}))

    def test_check_complete_error(self, mock_client, mock_plugin):
        mock_instance = mock.Mock(status="ERROR")
        mock_client().instances.get.return_value = mock_instance
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        exc = self.assertRaises(exception.ResourceInError,
                                trove.check_update_complete,
                                {"foo": "bar"})
        msg = "The last operation for the database instance failed"
        self.assertIn(msg, str(exc))

    def test_check_client_exceptions(self, mock_client, mock_plugin):
        mock_instance = mock.Mock(status="ACTIVE")
        mock_client().instances.get.return_value = mock_instance
        mock_plugin().is_client_exception.return_value = True
        mock_plugin().is_over_limit.side_effect = [True, False]
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        with mock.patch.object(trove, "_update_flavor") as mupdate:
            mupdate.side_effect = [Exception("test"),
                                   Exception("No change was requested "
                                             "because I'm testing")]
            self.assertFalse(trove.check_update_complete({"foo": "bar"}))
            self.assertFalse(trove.check_update_complete({"foo": "bar"}))
            self.assertEqual(2, mupdate.call_count)
            self.assertEqual(2, mock_plugin().is_client_exception.call_count)

    def test_check_complete_status(self, mock_client, mock_plugin):
        mock_instance = mock.Mock(status="RESIZING")
        mock_client().instances.get.return_value = mock_instance
        updates = {"foo": "bar"}
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertFalse(trove.check_update_complete(updates))

    def test_check_complete_name(self, mock_client, mock_plugin):
        mock_instance = mock.Mock(status="ACTIVE", name="mock_instance")
        mock_client().instances.get.return_value = mock_instance
        updates = {"name": "changed"}
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertFalse(trove.check_update_complete(updates))
        mock_instance.name = "changed"
        self.assertTrue(trove.check_update_complete(updates))
        mock_client().instances.edit.assert_called_once_with(mock_instance,
                                                             name="changed")

    def test_check_complete_databases(self, mock_client, mock_plugin):
        mock_instance = mock.Mock(status="ACTIVE", name="mock_instance")
        mock_client().instances.get.return_value = mock_instance
        updates = {
            'databases': [
                {'name': 'bar', "character_set": "ascii"},
                {'ACTION': 'CREATE', 'name': 'baz'},
                {'ACTION': 'DELETE', 'name': 'biff'}
            ]}
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertTrue(trove.check_update_complete(updates))
        mcreate = mock_client().databases.create
        mdelete = mock_client().databases.delete
        mcreate.assert_called_once_with(mock_instance, [{'name': 'baz'}])
        mdelete.assert_called_once_with(mock_instance, 'biff')

    def test_check_complete_users(self, mock_client, mock_plugin):
        mock_instance = mock.Mock(status="ACTIVE", name="mock_instance")
        mock_client().instances.get.return_value = mock_instance
        mock_plugin().is_client_exception.return_value = False
        mock_client().users.get.return_value = users.User(None, {
            "databases": [{
                "name": "bar"
            }, {
                "name": "buzz"
            }],
            "name": "baz"
            }, loaded=True)
        updates = {
            'users': [{
                'databases': ['bar', 'biff'],
                'name': 'baz',
                'password': 'changed'
            }, {
                'ACTION': 'CREATE',
                'databases': ['biff', 'bar'],
                'name': 'user2',
                'password': 'password'
            }, {
                'ACTION': 'DELETE',
                'name': 'deleted'
            }]}
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertTrue(trove.check_update_complete(updates))
        create_calls = [
            mock.call(mock_instance, [{'password': 'password',
                                       'databases': [{'name': 'biff'},
                                                     {'name': 'bar'}],
                                       'name': 'user2'}])
        ]
        delete_calls = [
            mock.call(mock_instance, 'deleted')
        ]
        mock_client().users.create.assert_has_calls(create_calls)
        mock_client().users.delete.assert_has_calls(delete_calls)
        self.assertEqual(1, mock_client().users.create.call_count)
        self.assertEqual(1, mock_client().users.delete.call_count)
        updateattr = mock_client().users.update_attributes
        updateattr.assert_called_once_with(
            mock_instance, 'baz', newuserattr={'password': 'changed'},
            hostname=mock.ANY)
        mock_client().users.grant.assert_called_once_with(
            mock_instance, 'baz', ['biff'])
        mock_client().users.revoke.assert_called_once_with(
            mock_instance, 'baz', ['buzz'])

    def test_check_complete_flavor(self, mock_client, mock_plugin):
        mock_instance = mock.Mock(status="ACTIVE", flavor={'id': 4567},
                                  name="mock_instance")
        mock_client().instances.get.return_value = mock_instance
        updates = {
            "flavor": 1234
        }
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertFalse(trove.check_update_complete(updates))
        mock_instance.status = "RESIZING"
        self.assertFalse(trove.check_update_complete(updates))
        mock_instance.status = "ACTIVE"
        mock_instance.flavor = {'id': 1234}
        self.assertTrue(trove.check_update_complete(updates))

    def test_check_complete_size(self, mock_client, mock_plugin):
        mock_instance = mock.Mock(status="ACTIVE", volume={'size': 24},
                                  name="mock_instance")
        mock_client().instances.get.return_value = mock_instance
        updates = {
            "size": 42
        }
        trove = dbinstance.Instance('test', self._rdef, self._stack)
        self.assertFalse(trove.check_update_complete(updates))
        mock_instance.status = "RESIZING"
        self.assertFalse(trove.check_update_complete(updates))
        mock_instance.status = "ACTIVE"
        mock_instance.volume = {'size': 42}
        self.assertTrue(trove.check_update_complete(updates))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from magnumclient import exceptions as mc_exc
from magnumclient.v1 import client as magnum_client
from oslo_config import cfg

from heat.common import exception
from heat.engine.clients import client_plugin
from heat.engine import constraints

CLIENT_NAME = 'magnum'


class MagnumClientPlugin(client_plugin.ClientPlugin):

    service_types = [CONTAINER] = ['container-infra']

    def _create(self):
        interface = self._get_client_option(CLIENT_NAME, 'endpoint_type')
        args = {
            'interface': interface,
            'service_type': self.CONTAINER,
            'session': self.context.keystone_session,
            'connect_retries': cfg.CONF.client_retry_limit,
            'region_name': self._get_region_name()
        }
        client = magnum_client.Client(**args)
        return client

    def is_not_found(self, ex):
        return isinstance(ex, mc_exc.NotFound)

    def is_over_limit(self, ex):
        return isinstance(ex, mc_exc.RequestEntityTooLarge)

    def is_conflict(self, ex):
        return isinstance(ex, mc_exc.Conflict)

    def _get_rsrc_name_or_id(self, value, entity, entity_msg):
        entity_client = getattr(self.client(), entity)
        try:
            return entity_client.get(value).uuid
        except mc_exc.NotFound:
            # Magnum cli will find the value either is name or id,
            # so no need to call list() here.
            raise exception.EntityNotFound(entity=entity_msg,
                                           name=value)

    def get_cluster_template(self, value):
        return self._get_rsrc_name_or_id(value, entity='cluster_templates',
                                         entity_msg='ClusterTemplate')


class ClusterTemplateConstraint(constraints.BaseCustomConstraint):

    resource_client_name = CLIENT_NAME
    resource_getter_name = 'get_cluster_template'



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.scaling import scalingutil as sc_util
from heat.tests import common


class TestCapacityChanges(common.HeatTestCase):
    # below:
    # n CHANGE_IN_CAPACITY (+up, -down)
    # b bounded
    # r rounded (+up, -down)
    # e EXACT_CAPACITY
    # p PERCENT_CHANGE_IN_CAPACITY
    # s MIN_ADJUSTMENT_STEP
    scenarios = [
        ('+n', dict(current=2, adjustment=3,
                    adjustment_type=sc_util.CHANGE_IN_CAPACITY,
                    min_adjustment_step=None,
                    minimum=0, maximum=10, expected=5)),
        ('-n', dict(current=6, adjustment=-2,
                    adjustment_type=sc_util.CHANGE_IN_CAPACITY,
                    min_adjustment_step=None,
                    minimum=0, maximum=5, expected=4)),
        ('+nb', dict(current=2, adjustment=8,
                     adjustment_type=sc_util.CHANGE_IN_CAPACITY,
                     min_adjustment_step=None,
                     minimum=0, maximum=5, expected=5)),
        ('-nb', dict(current=2, adjustment=-10,
                     adjustment_type=sc_util.CHANGE_IN_CAPACITY,
                     min_adjustment_step=None,
                     minimum=1, maximum=5, expected=1)),
        ('e', dict(current=2, adjustment=4,
                   adjustment_type=sc_util.EXACT_CAPACITY,
                   min_adjustment_step=None,
                   minimum=0, maximum=5, expected=4)),
        ('+eb', dict(current=2, adjustment=11,
                     adjustment_type=sc_util.EXACT_CAPACITY,
                     min_adjustment_step=None,
                     minimum=0, maximum=5, expected=5)),
        ('-eb', dict(current=4, adjustment=1,
                     adjustment_type=sc_util.EXACT_CAPACITY,
                     min_adjustment_step=None,
                     minimum=3, maximum=5, expected=3)),
        ('+p', dict(current=4, adjustment=50,
                    adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                    min_adjustment_step=None,
                    minimum=1, maximum=10, expected=6)),
        ('-p', dict(current=4, adjustment=-25,
                    adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                    min_adjustment_step=None,
                    minimum=1, maximum=10, expected=3)),
        ('+pb', dict(current=4, adjustment=100,
                     adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                     min_adjustment_step=None,
                     minimum=1, maximum=6, expected=6)),
        ('-pb', dict(current=6, adjustment=-50,
                     adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                     min_adjustment_step=None,
                     minimum=4, maximum=10, expected=4)),
        ('-p+r', dict(current=2, adjustment=-33,
                      adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                      min_adjustment_step=None,
                      minimum=0, maximum=10, expected=1)),
        ('+p+r', dict(current=1, adjustment=33,
                      adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                      min_adjustment_step=None,
                      minimum=0, maximum=10, expected=2)),
        ('-p-r', dict(current=2, adjustment=-66,
                      adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                      min_adjustment_step=None,
                      minimum=0, maximum=10, expected=1)),
        ('+p-r', dict(current=1, adjustment=225,
                      adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                      min_adjustment_step=None,
                      minimum=0, maximum=10, expected=3)),
        ('+ps', dict(current=1, adjustment=100,
                     adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                     min_adjustment_step=3,
                     minimum=0, maximum=10, expected=4)),
        ('+p+rs', dict(current=1, adjustment=33,
                       adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                       min_adjustment_step=2,
                       minimum=0, maximum=10, expected=3)),
        ('+p-rs', dict(current=1, adjustment=325,
                       adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                       min_adjustment_step=2,
                       minimum=0, maximum=10, expected=4)),
        ('-p-rs', dict(current=3, adjustment=-25,
                       adjustment_type=sc_util.PERCENT_CHANGE_IN_CAPACITY,
                       min_adjustment_step=2,
                       minimum=0, maximum=10, expected=1)),


    ]

    def test_calc(self):
        self.assertEqual(self.expected,
                         sc_util.calculate_new_capacity(
                             self.current, self.adjustment,
                             self.adjustment_type,
                             self.min_adjustment_step,
                             self.minimum, self.maximum))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _
from heat.engine.resources import signal_responder

LOG = logging.getLogger(__name__)


class BaseWaitConditionHandle(signal_responder.SignalResponder):
    """Base WaitConditionHandle resource.

    The main point of this class is to :
    - have no dependencies (so the instance can reference it)
    - create credentials to allow for signalling from the instance.
    - handle signals from the instance, validate and store result
    """
    properties_schema = {}

    WAIT_STATUSES = (
        STATUS_FAILURE,
        STATUS_SUCCESS,
    ) = (
        'FAILURE',
        'SUCCESS',
    )

    def _get_ec2_signed_url(self, signal_type=signal_responder.WAITCONDITION):
        stored = self.data().get('ec2_signed_url')
        if stored is not None:
            return stored
        url = super(BaseWaitConditionHandle,
                    self)._get_ec2_signed_url(signal_type)
        self.data_set('ec2_signed_url', url)
        return url

    def handle_create(self):
        super(BaseWaitConditionHandle, self).handle_create()
        self.resource_id_set(self._get_user_id())

    def _status_ok(self, status):
        return status in self.WAIT_STATUSES

    def _metadata_format_ok(self, metadata):
        if not isinstance(metadata, collections.abc.Mapping):
            return False
        if set(metadata) != set(self.METADATA_KEYS):
            return False
        return self._status_ok(metadata[self.STATUS])

    def normalise_signal_data(self, signal_data, latest_metadata):
        return signal_data

    def handle_signal(self, details=None):
        write_attempts = []

        def merge_signal_metadata(signal_data, latest_rsrc_metadata):
            signal_data = self.normalise_signal_data(signal_data,
                                                     latest_rsrc_metadata)

            if not self._metadata_format_ok(signal_data):
                LOG.info("Metadata failed validation for %s", self.name)
                raise ValueError(_("Metadata format invalid"))

            new_entry = signal_data.copy()
            unique_id = str(new_entry.pop(self.UNIQUE_ID))

            new_rsrc_metadata = latest_rsrc_metadata.copy()
            if unique_id in new_rsrc_metadata:
                LOG.info("Overwriting Metadata item for id %s!",
                         unique_id)
            new_rsrc_metadata.update({unique_id: new_entry})

            write_attempts.append(signal_data)
            return new_rsrc_metadata

        self.metadata_set(details, merge_metadata=merge_signal_metadata)

        data_written = write_attempts[-1]
        signal_reason = ('status:%s reason:%s' %
                         (data_written[self.STATUS],
                          data_written[self.REASON]))
        return signal_reason

    def get_status(self):
        """Return a list of the Status values for the handle signals."""
        return [v[self.STATUS]
                for v in self.metadata_get(refresh=True).values()]

    def get_status_reason(self, status):
        """Return a list of reasons associated with a particular status."""
        return [v[self.REASON]
                for v in self.metadata_get(refresh=True).values()
                if v[self.STATUS] == status]


class WaitConditionFailure(exception.Error):
    def __init__(self, wait_condition, handle):
        reasons = handle.get_status_reason(handle.STATUS_FAILURE)
        super(WaitConditionFailure, self).__init__(';'.join(reasons))


class WaitConditionTimeout(exception.Error):
    def __init__(self, wait_condition, handle):
        reasons = handle.get_status_reason(handle.STATUS_SUCCESS)
        vals = {'len': len(reasons),
                'count': wait_condition.properties[wait_condition.COUNT]}
        if reasons:
            vals['reasons'] = ';'.join(reasons)
            message = (_('%(len)d of %(count)d received - %(reasons)s') % vals)
        else:
            message = (_('%(len)d of %(count)d received') % vals)
        super(WaitConditionTimeout, self).__init__(message)



#
# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from keystoneauth1 import exceptions as keystone_exc
from oslo_config import cfg
import webob

from heat.common import auth_password
from heat.tests import common


cfg.CONF.import_opt('keystone_backend',
                    'heat.engine.clients.os.keystone.heat_keystoneclient')


EXPECTED_ENV_RESPONSE = {
    'HTTP_X_IDENTITY_STATUS': 'Confirmed',
    'HTTP_X_PROJECT_ID': 'tenant_id1',
    'HTTP_X_PROJECT_NAME': 'tenant_name1',
    'HTTP_X_USER_ID': 'user_id1',
    'HTTP_X_USER_DOMAIN_ID': 'domain1',
    'HTTP_X_USER_NAME': 'user_name1',
    'HTTP_X_ROLES': 'role1,role2',
    'HTTP_X_AUTH_TOKEN': 'lalalalalala',
}

TOKEN_V3_RESPONSE = {
    'version': 'v3',
    'project_id': 'tenant_id1',
    'project_name': 'tenant_name1',
    'user_id': 'user_id1',
    'username': 'user_name1',
    'service_catalog': None,
    'role_names': ['role1', 'role2'],
    'auth_token': 'lalalalalala',
    'user_domain_id': 'domain1'
}

TOKEN_V2_RESPONSE = {
    'version': 'v2',
    'tenant_id': 'tenant_id1',
    'tenant_name': 'tenant_name1',
    'user_id': 'user_id1',
    'service_catalog': None,
    'username': 'user_name1',
    'role_names': ['role1', 'role2'],
    'auth_token': 'lalalalalala',
    'user_domain_id': 'domain1'
}


class FakeAccessInfo(object):
    def __init__(self, **args):
        self.__dict__.update(args)


class FakeApp(object):
    """This represents a WSGI app protected by our auth middleware."""

    def __init__(self, expected_env=None):
        expected_env = expected_env or {}
        self.expected_env = dict(EXPECTED_ENV_RESPONSE)
        self.expected_env.update(expected_env)

    def __call__(self, env, start_response):
        """Assert that expected environment is present when finally called."""
        for k, v in self.expected_env.items():
            if env[k] != v:
                raise AssertionError('%s != %s' % (env[k], v))
        resp = webob.Response()
        resp.body = 'SUCCESS'.encode('latin-1')
        return resp(env, start_response)


class KeystonePasswordAuthProtocolTest(common.HeatTestCase):

    def setUp(self):
        super(KeystonePasswordAuthProtocolTest, self).setUp()
        self.config = {'auth_uri': 'http://keystone.test.com:5000'}
        self.app = FakeApp()
        self.middleware = auth_password.KeystonePasswordAuthProtocol(
            self.app, self.config)

    def _start_fake_response(self, status, headers):
        self.response_status = int(status.split(' ', 1)[0])
        self.response_headers = dict(headers)

    @mock.patch('keystoneauth1.identity.generic.Password')
    def test_valid_v2_request(self, mock_password):
        mock_auth = mock.MagicMock()
        mock_password.return_value = mock_auth

        self.patchobject(mock_auth, 'get_access',
                         return_value=FakeAccessInfo(**TOKEN_V2_RESPONSE))

        req = webob.Request.blank('/tenant_id1/')
        req.headers['X_AUTH_USER'] = 'user_name1'
        req.headers['X_AUTH_KEY'] = 'goodpassword'
        req.headers['X_AUTH_URL'] = self.config['auth_uri']
        req.headers['X_USER_DOMAIN_ID'] = 'domain1'
        self.middleware(req.environ, self._start_fake_response)
        mock_password.assert_called_once_with(auth_url=self.config['auth_uri'],
                                              password='goodpassword',
                                              project_id='tenant_id1',
                                              user_domain_id='domain1',
                                              username='user_name1')

    @mock.patch('keystoneauth1.identity.generic.Password')
    def test_valid_v3_request(self, mock_password):
        mock_auth = mock.MagicMock()
        mock_password.return_value = mock_auth

        self.patchobject(mock_auth, 'get_access',
                         return_value=FakeAccessInfo(**TOKEN_V3_RESPONSE))

        req = webob.Request.blank('/tenant_id1/')
        req.headers['X_AUTH_USER'] = 'user_name1'
        req.headers['X_AUTH_KEY'] = 'goodpassword'
        req.headers['X_AUTH_URL'] = self.config['auth_uri']
        req.headers['X_USER_DOMAIN_ID'] = 'domain1'
        self.middleware(req.environ, self._start_fake_response)
        mock_password.assert_called_once_with(auth_url=self.config['auth_uri'],
                                              password='goodpassword',
                                              project_id='tenant_id1',
                                              user_domain_id='domain1',
                                              username='user_name1')

    @mock.patch('keystoneauth1.identity.generic.Password')
    def test_request_with_bad_credentials(self, mock_password):
        mock_password.side_effect = keystone_exc.Unauthorized(401)

        req = webob.Request.blank('/tenant_id1/')
        req.headers['X_AUTH_USER'] = 'user_name1'
        req.headers['X_AUTH_KEY'] = 'badpassword'
        req.headers['X_AUTH_URL'] = self.config['auth_uri']
        req.headers['X_USER_DOMAIN_ID'] = 'domain1'
        self.middleware(req.environ, self._start_fake_response)
        self.assertEqual(401, self.response_status)
        mock_password.assert_called_once_with(auth_url=self.config['auth_uri'],
                                              password='badpassword',
                                              project_id='tenant_id1',
                                              user_domain_id='domain1',
                                              username='user_name1')

    def test_request_with_no_tenant_in_url_or_auth_headers(self):
        req = webob.Request.blank('/')
        self.middleware(req.environ, self._start_fake_response)
        self.assertEqual(401, self.response_status)
